from decimal import Decimal
from typing import List, Optional, Tuple, Union

from pydantic import Field, field_validator
from pydantic_core.core_schema import ValidationInfo

from hummingbot.core.data_type.common import MarketDict, OrderType, PositionMode, PriceType, TradeType
from hummingbot.strategy_v2.controllers.controller_base import ControllerBase, ControllerConfigBase
from hummingbot.strategy_v2.executors.data_types import ConnectorPair
from hummingbot.strategy_v2.executors.order_executor.data_types import ExecutionStrategy, OrderExecutorConfig
from hummingbot.strategy_v2.executors.position_executor.data_types import TrailingStop, TripleBarrierConfig
from hummingbot.strategy_v2.models.executor_actions import CreateExecutorAction, ExecutorAction, StopExecutorAction
from hummingbot.strategy_v2.models.executors import CloseType


class MarketMakingControllerConfigBase(ControllerConfigBase):
    """
    This class represents the base configuration for a market making controller.
    """
    controller_type: str = "market_making"
    connector_name: str = Field(
        default="binance_perpetual",
        json_schema_extra={
            "prompt": "Enter the connector name (e.g., binance_perpetual): ",
            "prompt_on_new": True}
    )
    trading_pair: str = Field(
        default="WLD-USDT",
        json_schema_extra={
            "prompt": "Enter the trading pair to trade on (e.g., WLD-USDT): ",
            "prompt_on_new": True}
    )
    buy_spreads: List[float] = Field(
        default="0.01,0.02",
        json_schema_extra={
            "prompt": "Enter a comma-separated list of buy spreads (e.g., '0.01, 0.02'): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    sell_spreads: List[float] = Field(
        default="0.01,0.02",
        json_schema_extra={
            "prompt": "Enter a comma-separated list of sell spreads (e.g., '0.01, 0.02'): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    buy_amounts_pct: Union[List[Decimal], None] = Field(
        default=None,
        json_schema_extra={
            "prompt": "Enter a comma-separated list of buy amounts as percentages (e.g., '50, 50'), or leave blank to distribute equally: ",
            "prompt_on_new": True, "is_updatable": True}
    )
    sell_amounts_pct: Union[List[Decimal], None] = Field(
        default=None,
        json_schema_extra={
            "prompt": "Enter a comma-separated list of sell amounts as percentages (e.g., '50, 50'), or leave blank to distribute equally: ",
            "prompt_on_new": True, "is_updatable": True}
    )
    executor_refresh_time: int = Field(
        default=60 * 5,
        json_schema_extra={
            "prompt": "Enter the refresh time in seconds for executors (e.g., 300 for 5 minutes): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    cooldown_time: int = Field(
        default=15,
        json_schema_extra={
            "prompt": "Enter the cooldown time in seconds between replacing an executor that traded (e.g., 15): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    leverage: int = Field(
        default=20,
        json_schema_extra={
            "prompt": "Enter the leverage to use for trading (e.g., 20 for 20x leverage). Set it to 1 for spot trading: ",
            "prompt_on_new": True}
    )
    position_mode: PositionMode = Field(
        default="HEDGE",
        json_schema_extra={"prompt": "Enter the position mode (HEDGE/ONEWAY): "}
    )
    # Triple Barrier Configuration
    stop_loss: Optional[Decimal] = Field(
        default=Decimal("0.03"), gt=0,
        json_schema_extra={
            "prompt": "Enter the stop loss (as a decimal, e.g., 0.03 for 3%): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    take_profit: Optional[Decimal] = Field(
        default=Decimal("0.02"), gt=0,
        json_schema_extra={
            "prompt": "Enter the take profit (as a decimal, e.g., 0.02 for 2%): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    time_limit: Optional[int] = Field(
        default=60 * 45, gt=0,
        json_schema_extra={
            "prompt": "Enter the time limit in seconds (e.g., 2700 for 45 minutes): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    take_profit_order_type: OrderType = Field(
        default=OrderType.LIMIT,
        json_schema_extra={
            "prompt": "Enter the order type for take profit (LIMIT/MARKET): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    trailing_stop: Optional[TrailingStop] = Field(
        default=None,
        json_schema_extra={
            "prompt": "Enter the trailing stop as activation_price,trailing_delta (e.g., 0.015,0.003): ",
            "prompt_on_new": True, "is_updatable": True},
    )
    # Position Management Configuration
    position_rebalance_threshold_pct: Decimal = Field(
        default=Decimal("0.05"),
        json_schema_extra={
            "prompt": "Enter the position rebalance threshold percentage (e.g., 0.05 for 5%): ",
            "prompt_on_new": True, "is_updatable": True}
    )
    skip_rebalance: bool = Field(default=False)

    @field_validator("trailing_stop", mode="before")
    @classmethod
    def parse_trailing_stop(cls, v):
        if isinstance(v, str):
            if v == "":
                return None
            activation_price, trailing_delta = v.split(",")
            return TrailingStop(activation_price=Decimal(activation_price), trailing_delta=Decimal(trailing_delta))
        return v

    @field_validator("time_limit", "stop_loss", "take_profit", "position_rebalance_threshold_pct", mode="before")
    @classmethod
    def validate_target(cls, v):
        if isinstance(v, str):
            if v == "":
                return None
            return Decimal(v)
        return v

    @field_validator('take_profit_order_type', mode="before")
    @classmethod
    def validate_order_type(cls, v) -> OrderType:
        if isinstance(v, OrderType):
            return v
        elif v is None:
            return OrderType.MARKET
        elif isinstance(v, str):
            cleaned_str = v.replace("OrderType.", "").upper()
            if cleaned_str in OrderType.__members__:
                return OrderType[cleaned_str]
        elif isinstance(v, int):
            try:
                return OrderType(v)
            except ValueError:
                pass
        raise ValueError(f"Invalid order type: {v}. Valid options are: {', '.join(OrderType.__members__)}")

    @field_validator('position_mode', mode="before")
    @classmethod
    def validate_position_mode(cls, v: str) -> PositionMode:
        if isinstance(v, str):
            if v.upper() in PositionMode.__members__:
                return PositionMode[v.upper()]
            raise ValueError(f"Invalid position mode: {v}. Valid options are: {', '.join(PositionMode.__members__)}")
        return v

    @field_validator('buy_spreads', 'sell_spreads', mode="before")
    @classmethod
    def parse_spreads(cls, v):
        if v is None:
            return []
        if isinstance(v, str):
            if v == "":
                return []
            return [float(x.strip()) for x in v.split(',')]
        return v

    @field_validator('buy_amounts_pct', 'sell_amounts_pct', mode="before")
    @classmethod
    def parse_and_validate_amounts(cls, v, validation_info: ValidationInfo):
        field_name = validation_info.field_name
        if v is None or v == "":
            spread_field = field_name.replace('amounts_pct', 'spreads')
            return [1 for _ in validation_info.data[spread_field]]
        if isinstance(v, str):
            return [float(x.strip()) for x in v.split(',')]
        elif isinstance(v, list) and len(v) != len(validation_info.data[field_name.replace('amounts_pct', 'spreads')]):
            raise ValueError(
                f"The number of {field_name} must match the number of {field_name.replace('amounts_pct', 'spreads')}.")
        return v

    @property
    def triple_barrier_config(self) -> TripleBarrierConfig:
        return TripleBarrierConfig(
            stop_loss=self.stop_loss,
            take_profit=self.take_profit,
            time_limit=self.time_limit,
            trailing_stop=self.trailing_stop,
            open_order_type=OrderType.LIMIT,  # Defaulting to LIMIT as is a Maker Controller
            take_profit_order_type=self.take_profit_order_type,
            stop_loss_order_type=OrderType.MARKET,  # Defaulting to MARKET as per requirement
            time_limit_order_type=OrderType.MARKET  # Defaulting to MARKET as per requirement
        )

    def get_spreads_and_amounts_in_quote(self, trade_type: TradeType) -> Tuple[List[float], List[float]]:
        buy_amounts_pct = getattr(self, 'buy_amounts_pct')
        sell_amounts_pct = getattr(self, 'sell_amounts_pct')

        # Calculate total percentages across buys and sells
        total_pct = sum(buy_amounts_pct) + sum(sell_amounts_pct)

        # Normalize amounts_pct based on total percentages
        if trade_type == TradeType.BUY:
            normalized_amounts_pct = [amt_pct / total_pct for amt_pct in buy_amounts_pct]
        else:  # TradeType.SELL
            normalized_amounts_pct = [amt_pct / total_pct for amt_pct in sell_amounts_pct]

        spreads = getattr(self, f'{trade_type.name.lower()}_spreads')
        return spreads, [amt_pct * self.total_amount_quote for amt_pct in normalized_amounts_pct]

    def get_required_base_amount(self, reference_price: Decimal) -> Decimal:
        """
        Get the required base asset amount for sell orders.
        """
        _, sell_amounts_quote = self.get_spreads_and_amounts_in_quote(TradeType.SELL)
        total_sell_amount_quote = sum(sell_amounts_quote)
        return total_sell_amount_quote / reference_price

    def update_markets(self, markets: MarketDict) -> MarketDict:
        return markets.add_or_update(self.connector_name, self.trading_pair)


class MarketMakingControllerBase(ControllerBase):
    """
    This class represents the base class for a market making controller.
    """

    def __init__(self, config: MarketMakingControllerConfigBase, *args, **kwargs):
        super().__init__(config, *args, **kwargs)
        self.config = config
        self.market_data_provider.initialize_rate_sources([ConnectorPair(
            connector_name=config.connector_name, trading_pair=config.trading_pair)])

    def determine_executor_actions(self) -> List[ExecutorAction]:
        """
        Determine actions based on the provided executor handler report.
        """
        actions = []
        actions.extend(self.create_actions_proposal())
        actions.extend(self.stop_actions_proposal())
        return actions

    def create_actions_proposal(self) -> List[ExecutorAction]:
        """
        Create actions proposal based on the current state of the controller.
        """
        create_actions = []

        # Check if we need to rebalance position first
        position_rebalance_action = self.check_position_rebalance()
        if position_rebalance_action:
            create_actions.append(position_rebalance_action)

        # Create normal market making levels
        levels_to_execute = self.get_levels_to_execute()
        for level_id in levels_to_execute:
            price, amount = self.get_price_and_amount(level_id)
            executor_config = self.get_executor_config(level_id, price, amount)
            if executor_config is not None:
                create_actions.append(CreateExecutorAction(
                    controller_id=self.config.id,
                    executor_config=executor_config
                ))
        return create_actions

    def get_levels_to_execute(self) -> List[str]:
        working_levels = self.filter_executors(
            executors=self.executors_info,
            filter_func=lambda x: x.is_active or (x.close_type == CloseType.STOP_LOSS and self.market_data_provider.time() - x.close_timestamp < self.config.cooldown_time)
        )
        working_levels_ids = [executor.custom_info["level_id"] for executor in working_levels]
        return self.get_not_active_levels_ids(working_levels_ids)

    def stop_actions_proposal(self) -> List[ExecutorAction]:
        """
        Create a list of actions to stop the executors based on order refresh and early stop conditions.
        """
        stop_actions = []
        stop_actions.extend(self.executors_to_refresh())
        stop_actions.extend(self.executors_to_early_stop())
        return stop_actions

    def executors_to_refresh(self) -> List[ExecutorAction]:
        executors_to_refresh = self.filter_executors(
            executors=self.executors_info,
            filter_func=lambda x: not x.is_trading and x.is_active and self.market_data_provider.time() - x.timestamp > self.config.executor_refresh_time)

        return [StopExecutorAction(
            controller_id=self.config.id,
            executor_id=executor.id) for executor in executors_to_refresh]

    def executors_to_early_stop(self) -> List[ExecutorAction]:
        """
        Get the executors to early stop based on the current state of market data. This method can be overridden to
        implement custom behavior.
        """
        return []

    async def update_processed_data(self):
        """
        Update the processed data for the controller. This method should be reimplemented to modify the reference price
        and spread multiplier based on the market data. By default, it will update the reference price as mid price and
        the spread multiplier as 1.
        """
        reference_price = self.market_data_provider.get_price_by_type(self.config.connector_name,
                                                                      self.config.trading_pair, PriceType.MidPrice)
        self.processed_data = {"reference_price": Decimal(reference_price), "spread_multiplier": Decimal("1")}

    def get_executor_config(self, level_id: str, price: Decimal, amount: Decimal):
        """
        Get the executor config for a given level id.
        """
        raise NotImplementedError

    def get_price_and_amount(self, level_id: str) -> Tuple[Decimal, Decimal]:
        """
        Get the spread and amount in quote for a given level id.
        """
        level = self.get_level_from_level_id(level_id)
        trade_type = self.get_trade_type_from_level_id(level_id)
        spreads, amounts_quote = self.config.get_spreads_and_amounts_in_quote(trade_type)
        reference_price = Decimal(self.processed_data["reference_price"])
        spread_in_pct = Decimal(spreads[int(level)]) * Decimal(self.processed_data["spread_multiplier"])
        side_multiplier = Decimal("-1") if trade_type == TradeType.BUY else Decimal("1")
        order_price = reference_price * (1 + side_multiplier * spread_in_pct)
        return order_price, Decimal(amounts_quote[int(level)]) / order_price

    def get_level_id_from_side(self, trade_type: TradeType, level: int) -> str:
        """
        Get the level id based on the trade type and the level.
        """
        return f"{trade_type.name.lower()}_{level}"

    def get_trade_type_from_level_id(self, level_id: str) -> TradeType:
        return TradeType.BUY if level_id.startswith("buy") else TradeType.SELL

    def get_level_from_level_id(self, level_id: str) -> int:
        return int(level_id.split('_')[1])

    def get_not_active_levels_ids(self, active_levels_ids: List[str]) -> List[str]:
        """
        Get the levels to execute based on the current state of the controller.
        """
        buy_ids_missing = [self.get_level_id_from_side(TradeType.BUY, level) for level in range(len(self.config.buy_spreads))
                           if self.get_level_id_from_side(TradeType.BUY, level) not in active_levels_ids]
        sell_ids_missing = [self.get_level_id_from_side(TradeType.SELL, level) for level in range(len(self.config.sell_spreads))
                            if self.get_level_id_from_side(TradeType.SELL, level) not in active_levels_ids]
        return buy_ids_missing + sell_ids_missing

    def check_position_rebalance(self) -> Optional[CreateExecutorAction]:
        """
        Check if position needs rebalancing and create OrderExecutor to acquire missing base asset.
        Only applies to spot trading (not perpetual contracts).
        """
        # Skip position rebalancing for perpetual contracts
        if "_perpetual" in self.config.connector_name or "reference_price" not in self.processed_data or self.config.skip_rebalance:
            return None

        active_rebalance = self.filter_executors(
            executors=self.executors_info,
            filter_func=lambda x: x.is_active and x.custom_info.get("level_id") == "position_rebalance"
        )
        if len(active_rebalance) > 0:
            # If there's already an active rebalance executor, skip rebalancing
            return None

        required_base_amount = self.config.get_required_base_amount(Decimal(self.processed_data["reference_price"]))
        current_base_amount = self.get_current_base_position()

        # Calculate the difference
        base_amount_diff = required_base_amount - current_base_amount

        # Check if difference exceeds threshold
        threshold_amount = required_base_amount * self.config.position_rebalance_threshold_pct

        if abs(base_amount_diff) > threshold_amount:
            # We need to rebalance
            if base_amount_diff > 0:
                # Need to buy more base asset
                return self.create_position_rebalance_order(TradeType.BUY, abs(base_amount_diff))
            else:
                # Need to sell base asset (unlikely for market making but possible)
                return self.create_position_rebalance_order(TradeType.SELL, abs(base_amount_diff))

        return None

    def get_current_base_position(self) -> Decimal:
        """
        Get current base asset position from positions held.
        """
        total_base_amount = Decimal("0")

        for position in self.positions_held:
            if (position.connector_name == self.config.connector_name and
                    position.trading_pair == self.config.trading_pair):
                # Calculate net base position
                if position.side == TradeType.BUY:
                    total_base_amount += position.amount
                else:  # SELL position
                    total_base_amount -= position.amount

        return total_base_amount

    def create_position_rebalance_order(self, side: TradeType, amount: Decimal) -> CreateExecutorAction:
        """
        Create an OrderExecutor to rebalance position.
        """
        reference_price = self.processed_data["reference_price"]

        # Use market price for quick execution
        order_config = OrderExecutorConfig(
            timestamp=self.market_data_provider.time(),
            connector_name=self.config.connector_name,
            trading_pair=self.config.trading_pair,
            execution_strategy=ExecutionStrategy.MARKET,
            side=side,
            amount=amount,
            price=reference_price,  # Will be ignored for market orders
            level_id="position_rebalance",
        )

        return CreateExecutorAction(
            controller_id=self.config.id,
            executor_config=order_config
        )



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.binance.com"
HEALTH_CHECK_ENDPOINT = "/api/v3/ping"
CANDLES_ENDPOINT = "/api/v3/klines"

WSS_URL = "wss://stream.binance.com:9443/ws"

INTERVALS = bidict({
    "1s": "1s",
    "1m": "1m",
    "3m": "3m",
    "5m": "5m",
    "15m": "15m",
    "30m": "30m",
    "1h": "1h",
    "2h": "2h",
    "4h": "4h",
    "6h": "6h",
    "8h": "8h",
    "12h": "12h",
    "1d": "1d",
    "3d": "3d",
    "1w": "1w",
    "1M": "1M"
})
MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1000
REQUEST_WEIGHT = "REQUEST_WEIGHT"

RATE_LIMITS = [
    RateLimit(REQUEST_WEIGHT, limit=6000, time_interval=60),
    RateLimit(CANDLES_ENDPOINT, limit=1200, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)]),
    RateLimit(HEALTH_CHECK_ENDPOINT, limit=1200, time_interval=60, linked_limits=[LinkedLimitWeightPair("raw", 1)])]



import platform
import threading
from typing import TYPE_CHECKING

from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.strategy.script_strategy_base import ScriptStrategyBase

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class StopCommand:
    def stop(self,  # type: HummingbotApplication
             skip_order_cancellation: bool = False):
        if threading.current_thread() != threading.main_thread():
            self.ev_loop.call_soon_threadsafe(self.stop, skip_order_cancellation)
            return
        safe_ensure_future(self.stop_loop(skip_order_cancellation), loop=self.ev_loop)

    async def stop_loop(self,  # type: HummingbotApplication
                        skip_order_cancellation: bool = False):
        self.logger().info("stop command initiated.")
        self.notify("\nWinding down...")

        # Restore App Nap on macOS.
        if platform.system() == "Darwin":
            import appnope
            appnope.nap()

        # Handle script strategy specific cleanup first
        if self.trading_core.strategy and isinstance(self.trading_core.strategy, ScriptStrategyBase):
            await self.trading_core.strategy.on_stop()

        # Stop strategy if running
        if self.trading_core._strategy_running:
            await self.trading_core.stop_strategy()

        # Cancel outstanding orders
        if not skip_order_cancellation:
            await self.trading_core.cancel_outstanding_orders()

        # Remove all connectors
        connector_names = list(self.trading_core.connectors.keys())
        for name in connector_names:
            try:
                self.trading_core.remove_connector(name)
            except Exception as e:
                self.logger().error(f"Error stopping connector {name}: {e}")

        # Stop clock if running
        if self.trading_core._is_running:
            await self.trading_core.stop_clock()

        # Stop markets recorder
        if self.trading_core.markets_recorder:
            self.trading_core.markets_recorder.stop()
            self.trading_core.markets_recorder = None

        # Clear strategy references
        self.trading_core.strategy = None
        self.trading_core.strategy_name = None
        self.trading_core.strategy_config_map = None
        self.trading_core._strategy_file_name = None
        self.trading_core._config_source = None
        self.trading_core._config_data = None

        self.notify("Hummingbot stopped.")



from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import TYPE_CHECKING, Dict, Any
if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication

from hummingbot.client.ui.custom_widgets import CustomTextArea


class TabBase(metaclass=ABCMeta):
    """
    Defines functions needed to be implemented by all tab classes.
    """

    @classmethod
    @abstractmethod
    def get_command_name(cls) -> str:
        """
        Returns a command name for the tab, once issued a new tab will be created. The command name will also appear on
        auto complete list of commands.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def get_command_help_message(cls) -> str:
        """
        Returns a help message to describe what the command does.
        """

        raise NotImplementedError

    @classmethod
    @abstractmethod
    def get_command_arguments(cls) -> Dict[str, Dict[str, Any]]:
        """
        Returns a dictionary of command argument and all its properties. See hummingbot.client.ui.parser for examples.
        """

        raise NotImplementedError

    @classmethod
    @abstractmethod
    async def display(cls, output_field: CustomTextArea, hummingbot: "HummingbotApplication", **kwargs):
        """
        Displays message on the tab
        :param output_field: The output pane for the tab messages
        :param hummingbot: The current running Hummingbot application including strategy, connectors and all
        other application properties
        :param **kargs: All the command arguments defined in get_command_arguments method will be supplied here
        """

        raise NotImplementedError



import asyncio
import logging
from decimal import Decimal
from typing import Dict, Optional, Set

import pandas as pd

from hummingbot.core.gateway.gateway_http_client import GatewayHttpClient
from hummingbot.core.network_base import NetworkBase
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger


class WalletTrackerDataFeed(NetworkBase):
    dex_logger: Optional[HummingbotLogger] = None
    _gateway_client = None

    @property
    def gateway_client(self):
        if self._gateway_client is None:
            self._gateway_client = GatewayHttpClient.get_instance()
        return self._gateway_client

    def __init__(
        self,
        chain: str,
        network: str,
        wallets: Set[str],
        tokens: Set[str],
        update_interval: float = 1.0,
    ) -> None:
        super().__init__()
        self._ev_loop = asyncio.get_event_loop()
        self._chain = chain
        self._network = network
        self._tokens = tokens
        self._wallet_balances: Dict[str, Dict[str, float]] = {wallet: {} for wallet in wallets}
        self._update_interval = update_interval
        self.fetch_data_loop_task: Optional[asyncio.Task] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls.dex_logger is None:
            cls.dex_logger = logging.getLogger(__name__)
        return cls.dex_logger

    @property
    def name(self) -> str:
        return f"WalletTrackerDataFeed[{self.chain}-{self.network}]"

    @property
    def chain(self) -> str:
        return self._chain

    @property
    def network(self) -> str:
        return self._network

    @property
    def tokens(self) -> Set[str]:
        return self._tokens

    @property
    def wallet_balances(self) -> Dict[str, Dict[str, float]]:
        return self._wallet_balances

    @property
    def wallet_balances_df(self) -> pd.DataFrame:
        return pd.DataFrame(self._wallet_balances).T

    def is_ready(self) -> bool:
        return all(len(wallet_balances) > 0 for wallet_balances in self._wallet_balances.values())

    async def check_network(self) -> NetworkStatus:
        is_gateway_online = await self.gateway_client.ping_gateway()
        if not is_gateway_online:
            self.logger().warning("Gateway is not online. Please check your gateway connection.")
        return NetworkStatus.CONNECTED if is_gateway_online else NetworkStatus.NOT_CONNECTED

    async def start_network(self) -> None:
        await self.stop_network()
        self.fetch_data_loop_task = safe_ensure_future(self._fetch_data_loop())

    async def stop_network(self) -> None:
        if self.fetch_data_loop_task is not None:
            self.fetch_data_loop_task.cancel()
            self.fetch_data_loop_task = None

    async def _fetch_data_loop(self) -> None:
        while True:
            try:
                await self._fetch_data()
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(
                    f"Error getting data from {self.name}"
                    f"Check network connection. Error: {e}",
                )
            await self._async_sleep(self._update_interval)

    async def _fetch_data(self) -> None:
        wallet_balances_tasks = [
            asyncio.create_task(self._update_balances_by_wallet(wallet))
            for wallet in self._wallet_balances.keys()
        ]
        await asyncio.gather(*wallet_balances_tasks)

    async def _update_balances_by_wallet(self, wallet: str) -> None:
        data = await self.gateway_client.get_balances(
            self.chain,
            self.network,
            wallet,
            list(self._tokens)
        )
        self._wallet_balances[wallet] = {token: Decimal(balance) for token, balance in data['balances'].items()}

    @staticmethod
    async def _async_sleep(delay: float) -> None:
        """Used to mock in test cases."""
        await asyncio.sleep(delay)



from typing import Dict, Optional

from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType


class MexcOrderBook(OrderBook):

    @classmethod
    def snapshot_message_from_exchange(cls,
                                       msg: Dict[str, any],
                                       timestamp: float,
                                       metadata: Optional[Dict] = None) -> OrderBookMessage:
        """
        Creates a snapshot message with the order book snapshot message
        :param msg: the response from the exchange when requesting the order book snapshot
        :param timestamp: the snapshot timestamp
        :param metadata: a dictionary with extra information to add to the snapshot data
        :return: a snapshot message with the snapshot information received from the exchange
        """
        if metadata:
            msg.update(metadata)
        return OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": msg["trading_pair"],
            "update_id": msg["lastUpdateId"],
            "bids": msg["bids"],
            "asks": msg["asks"]
        }, timestamp=float(timestamp))

    @classmethod
    def diff_message_from_exchange(cls,
                                   msg: Dict[str, any],
                                   timestamp: Optional[float] = None,
                                   metadata: Optional[Dict] = None) -> OrderBookMessage:
        """
        Creates a diff message with the changes in the order book received from the exchange
        :param msg: the changes in the order book
        :param timestamp: the timestamp of the difference
        :param metadata: a dictionary with extra information to add to the difference data
        :return: a diff message with the changes in the order book notified by the exchange
        """
        if metadata:
            msg.update(metadata)
        return OrderBookMessage(OrderBookMessageType.DIFF, {
            "trading_pair": msg["trading_pair"],
            "update_id": timestamp,
            "bids": [[i['price'], i['quantity']] for i in msg['publicAggreDepths'].get("bids", [])],
            "asks": [[i['price'], i['quantity']] for i in msg['publicAggreDepths'].get("asks", [])],
        }, timestamp=float(timestamp) * 1e-3)

    @classmethod
    def trade_message_from_exchange(cls,
                                    msg: Dict[str, any],
                                    timestamp: Optional[float] = None,
                                    metadata: Optional[Dict] = None):
        """
        Creates a trade message with the information from the trade event sent by the exchange
        :param msg: the trade event details sent by the exchange
        :param timestamp: the timestamp of the difference
        :param metadata: a dictionary with extra information to add to trade message
        :return: a trade message with the details of the trade as provided by the exchange
        """
        if metadata:
            msg.update(metadata)
        ts = timestamp
        return OrderBookMessage(OrderBookMessageType.TRADE, {
            "trading_pair": msg["trading_pair"],
            "trade_type": float(TradeType.SELL.value) if msg["tradeType"] == 2 else float(TradeType.BUY.value),
            "trade_id": msg["time"],
            "update_id": ts,
            "price": msg["price"],
            "amount": msg["quantity"]
        }, timestamp=float(ts) * 1e-3)



import asyncio
import copy
import logging
import math
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Any, AsyncIterable, Callable, Dict, List, Optional, Tuple

from async_timeout import timeout

from hummingbot.connector.client_order_tracker import ClientOrderTracker
from hummingbot.connector.constants import MINUTE, TWELVE_HOURS, s_decimal_0, s_decimal_NaN
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import get_new_client_order_id
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.cancellation_result import CancellationResult
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_tracker import OrderBookTracker
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee
from hummingbot.core.data_type.user_stream_tracker import UserStreamTracker
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future, safe_gather
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.logger import HummingbotLogger


class ExchangePyBase(ExchangeBase, ABC):
    _logger = None

    SHORT_POLL_INTERVAL = 5.0
    LONG_POLL_INTERVAL = 120.0
    TRADING_RULES_INTERVAL = 30 * MINUTE
    TRADING_FEES_INTERVAL = TWELVE_HOURS
    TICK_INTERVAL_LIMIT = 60.0

    def __init__(self,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100")):
        super().__init__(balance_asset_limit)

        self._last_poll_timestamp = 0
        self._last_timestamp = 0
        self._trading_rules = {}
        self._trading_fees = {}

        self._status_polling_task: Optional[asyncio.Task] = None
        self._user_stream_tracker_task: Optional[asyncio.Task] = None
        self._user_stream_event_listener_task: Optional[asyncio.Task] = None
        self._trading_rules_polling_task: Optional[asyncio.Task] = None
        self._trading_fees_polling_task: Optional[asyncio.Task] = None
        self._lost_orders_update_task: Optional[asyncio.Task] = None

        self._time_synchronizer = TimeSynchronizer()
        self._throttler = AsyncThrottler(
            rate_limits=self.rate_limits_rules,
            limits_share_percentage=rate_limits_share_pct)
        self._poll_notifier = asyncio.Event()

        # init Auth and Api factory
        self._auth: AuthBase = self.authenticator
        self._web_assistants_factory: WebAssistantsFactory = self._create_web_assistants_factory()

        # init OrderBook Data Source and Tracker
        self._orderbook_ds: OrderBookTrackerDataSource = self._create_order_book_data_source()
        self._set_order_book_tracker(OrderBookTracker(
            data_source=self._orderbook_ds,
            trading_pairs=self.trading_pairs,
            domain=self.domain))

        # init UserStream Data Source and Tracker
        self._user_stream_tracker = self._create_user_stream_tracker()

        self._order_tracker: ClientOrderTracker = self._create_order_tracker()

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(HummingbotLogger.logger_name_for_class(cls))
        return cls._logger

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def authenticator(self) -> AuthBase:
        raise NotImplementedError

    @property
    @abstractmethod
    def rate_limits_rules(self) -> List[RateLimit]:
        raise NotImplementedError

    @property
    @abstractmethod
    def domain(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def client_order_id_max_length(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def client_order_id_prefix(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def trading_rules_request_path(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def trading_pairs_request_path(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def check_network_request_path(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def trading_pairs(self) -> List[str]:
        raise NotImplementedError

    @property
    @abstractmethod
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        raise NotImplementedError

    @property
    @abstractmethod
    def is_trading_required(self) -> bool:
        raise NotImplementedError

    @property
    def order_books(self) -> Dict[str, OrderBook]:
        return self.order_book_tracker.order_books

    @property
    def in_flight_orders(self) -> Dict[str, InFlightOrder]:
        return self._order_tracker.active_orders

    @property
    def trading_rules(self) -> Dict[str, TradingRule]:
        return self._trading_rules

    @property
    def limit_orders(self) -> List[LimitOrder]:
        return [in_flight_order.to_limit_order() for in_flight_order in self.in_flight_orders.values()]

    @property
    def status_dict(self) -> Dict[str, bool]:
        return {
            "symbols_mapping_initialized": self.trading_pair_symbol_map_ready(),
            "order_books_initialized": self.order_book_tracker.ready,
            "account_balance": not self.is_trading_required or len(self._account_balances) > 0,
            "trading_rule_initialized": len(self._trading_rules) > 0 if self.is_trading_required else True,
            "user_stream_initialized": self._is_user_stream_initialized(),
        }

    @property
    def ready(self) -> bool:
        """
        Returns True if the connector is ready to operate (all connections established with the exchange). If it is
        not ready it returns False.
        """
        return all(self.status_dict.values())

    @property
    def name_cap(self) -> str:
        return self.name.capitalize()

    @property
    def tracking_states(self) -> Dict[str, any]:
        """
        Returns a dictionary associating current active orders client id to their JSON representation
        """
        return {key: value.to_json() for key, value in self._order_tracker.all_updatable_orders.items()}

    @abstractmethod
    def supported_order_types(self) -> List[OrderType]:
        raise NotImplementedError

    @abstractmethod
    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception) -> bool:
        raise NotImplementedError

    @abstractmethod
    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        raise NotImplementedError

    @abstractmethod
    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        raise NotImplementedError

    # === Price logic ===

    def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:
        """
        Used by quantize_order_price() in _create_order()
        Returns a price step, a minimum price increment for a given trading pair.

        :param trading_pair: the trading pair to check for market conditions
        :param price: the starting point price
        """
        trading_rule = self._trading_rules[trading_pair]
        return Decimal(trading_rule.min_price_increment)

    def get_order_size_quantum(self, trading_pair: str, order_size: Decimal) -> Decimal:
        """
        Used by quantize_order_price() in _create_order()
        Returns an order amount step, a minimum amount increment for a given trading pair.

        :param trading_pair: the trading pair to check for market conditions
        :param order_size: the starting point order price
        """
        trading_rule = self._trading_rules[trading_pair]
        return Decimal(trading_rule.min_base_amount_increment)

    def get_order_book(self, trading_pair: str) -> OrderBook:
        """
        Returns the current order book for a particular market

        :param trading_pair: the pair of tokens for which the order book should be retrieved
        """
        if trading_pair not in self.order_book_tracker.order_books:
            raise ValueError(f"No order book exists for '{trading_pair}'.")
        return self.order_book_tracker.order_books[trading_pair]

    def tick(self, timestamp: float):
        """
        Includes the logic that has to be processed every time a new tick happens in the bot. Particularly it enables
        the execution of the status update polling loop using an event.
        """
        poll_interval = self._get_poll_interval(timestamp=timestamp)
        last_tick = int(self._last_timestamp / poll_interval)
        current_tick = int(timestamp / poll_interval)
        if current_tick > last_tick:
            self._poll_notifier.set()
        self._last_timestamp = timestamp

    # === Orders placing ===

    def buy(self,
            trading_pair: str,
            amount: Decimal,
            order_type=OrderType.LIMIT,
            price: Decimal = s_decimal_NaN,
            **kwargs) -> str:
        """
        Creates a promise to create a buy order using the parameters

        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price

        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=True,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        safe_ensure_future(self._create_order(
            trade_type=TradeType.BUY,
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return order_id

    def sell(self,
             trading_pair: str,
             amount: Decimal,
             order_type: OrderType = OrderType.LIMIT,
             price: Decimal = s_decimal_NaN,
             **kwargs) -> str:
        """
        Creates a promise to create a sell order using the parameters.
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=False,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        safe_ensure_future(self._create_order(
            trade_type=TradeType.SELL,
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return order_id

    def get_fee(self,
                base_currency: str,
                quote_currency: str,
                order_type: OrderType,
                order_side: TradeType,
                amount: Decimal,
                price: Decimal = s_decimal_NaN,
                is_maker: Optional[bool] = None) -> AddedToCostTradeFee:
        """
        Calculates the fee to pay based on the fee information provided by the exchange for
        the account and the token pair. If exchange info is not available it calculates the estimated
        fee an order would pay based on the connector configuration.

        :param base_currency: the order base currency
        :param quote_currency: the order quote currency
        :param order_type: the type of order (MARKET, LIMIT, LIMIT_MAKER)
        :param order_side: if the order is for buying or selling
        :param amount: the order amount
        :param price: the order price
        :param is_maker: True if the order is a maker order, False if it is a taker order

        :return: the calculated or estimated fee
        """
        return self._get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)

    def cancel(self, trading_pair: str, client_order_id: str):
        """
        Creates a promise to cancel an order in the exchange

        :param trading_pair: the trading pair the order to cancel operates with
        :param client_order_id: the client id of the order to cancel

        :return: the client id of the order to cancel
        """
        safe_ensure_future(self._execute_cancel(trading_pair, client_order_id))
        return client_order_id

    async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:
        """
        Cancels all currently active orders. The cancellations are performed in parallel tasks.

        :param timeout_seconds: the maximum time (in seconds) the cancel logic should run

        :return: a list of CancellationResult instances, one for each of the orders to be cancelled
        """
        incomplete_orders = [o for o in self.in_flight_orders.values() if not o.is_done]
        tasks = [self._execute_cancel(o.trading_pair, o.client_order_id) for o in incomplete_orders]
        order_id_set = set([o.client_order_id for o in incomplete_orders])
        successful_cancellations = []

        try:
            async with timeout(timeout_seconds):
                cancellation_results = await safe_gather(*tasks, return_exceptions=True)
                for cr in cancellation_results:
                    if isinstance(cr, Exception):
                        continue
                    client_order_id = cr
                    if client_order_id is not None:
                        order_id_set.remove(client_order_id)
                        successful_cancellations.append(CancellationResult(client_order_id, True))
        except Exception:
            self.logger().network(
                "Unexpected error cancelling orders.",
                exc_info=True,
                app_warning_msg="Failed to cancel order. Check API key and network connection."
            )
        failed_cancellations = [CancellationResult(oid, False) for oid in order_id_set]
        return successful_cancellations + failed_cancellations

    async def _create_order(self,
                            trade_type: TradeType,
                            order_id: str,
                            trading_pair: str,
                            amount: Decimal,
                            order_type: OrderType,
                            price: Optional[Decimal] = None,
                            **kwargs):
        """
        Creates an order in the exchange using the parameters to configure it

        :param trade_type: the side of the order (BUY of SELL)
        :param order_id: the id that should be assigned to the order (the client id)
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        """
        trading_rule = self._trading_rules[trading_pair]

        if order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER]:
            price = self.quantize_order_price(trading_pair, price)
        quantized_amount = self.quantize_order_amount(trading_pair=trading_pair, amount=amount)

        self.start_tracking_order(
            order_id=order_id,
            exchange_order_id=None,
            trading_pair=trading_pair,
            order_type=order_type,
            trade_type=trade_type,
            price=price,
            amount=quantized_amount,
            **kwargs,
        )
        order = self._order_tracker.active_orders[order_id]
        if not price or price.is_nan() or price == s_decimal_0:
            current_price: Decimal = self.get_price(trading_pair, False)
            notional_size = current_price * quantized_amount
        else:
            notional_size = price * quantized_amount

        if order_type not in self.supported_order_types():
            self.logger().error(f"{order_type} is not in the list of supported order types")
            self._update_order_after_failure(
                order_id=order_id, trading_pair=trading_pair,
                exception=ValueError(f"{order_type} is not in the list of supported order types"))
            return

        elif quantized_amount < trading_rule.min_order_size:
            self._update_order_after_failure(
                order_id=order_id, trading_pair=trading_pair,
                exception=ValueError(f"Order amount {amount} is lower than minimum order size {trading_rule.min_order_size} "
                                     f"for the pair {trading_pair}. The order will not be created."))
            return

        elif notional_size < trading_rule.min_notional_size:
            self._update_order_after_failure(
                order_id=order_id, trading_pair=trading_pair,
                exception=ValueError(f"Order notional {notional_size} is lower than minimum notional size {trading_rule.min_notional_size}"
                                     f" for the pair {trading_pair}. The order will not be created."))
            return
        try:
            await self._place_order_and_process_update(order=order, **kwargs,)

        except asyncio.CancelledError:
            raise
        except Exception as ex:
            self._on_order_failure(
                order_id=order_id,
                trading_pair=trading_pair,
                amount=quantized_amount,
                trade_type=trade_type,
                order_type=order_type,
                price=price,
                exception=ex,
                **kwargs,
            )

    async def _place_order_and_process_update(self, order: InFlightOrder, **kwargs) -> str:
        exchange_order_id, update_timestamp = await self._place_order(
            order_id=order.client_order_id,
            trading_pair=order.trading_pair,
            amount=order.amount,
            trade_type=order.trade_type,
            order_type=order.order_type,
            price=order.price,
            **kwargs,
        )

        order_update: OrderUpdate = OrderUpdate(
            client_order_id=order.client_order_id,
            exchange_order_id=str(exchange_order_id),
            trading_pair=order.trading_pair,
            update_timestamp=update_timestamp,
            new_state=OrderState.OPEN,
        )
        self._order_tracker.process_order_update(order_update)

        return exchange_order_id

    def _on_order_failure(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Optional[Decimal],
        exception: Exception,
        **kwargs,
    ):
        self.logger().network(
            f"Error submitting {trade_type.name.lower()} {order_type.name.upper()} order to {self.name_cap} for "
            f"{amount} {trading_pair} {price}.",
            exc_info=True,
            app_warning_msg=f"Failed to submit {trade_type.name.upper()} order to {self.name_cap}. Check API key and network connection."
        )
        self._update_order_after_failure(order_id=order_id, trading_pair=trading_pair, exception=exception)

    def _update_order_after_failure(self, order_id: str, trading_pair: str, exception: Optional[Exception] = None):
        misc_updates = {}
        if exception:
            misc_updates['error_message'] = str(exception)
            misc_updates['error_type'] = exception.__class__.__name__

        order_update: OrderUpdate = OrderUpdate(
            client_order_id=order_id,
            trading_pair=trading_pair,
            update_timestamp=self.current_timestamp,
            new_state=OrderState.FAILED,
            misc_updates=misc_updates
        )
        self._order_tracker.process_order_update(order_update)

    async def _execute_order_cancel(self, order: InFlightOrder) -> Optional[str]:
        try:
            cancelled = await self._execute_order_cancel_and_process_update(order=order)
            if cancelled:
                return order.client_order_id
        except asyncio.CancelledError:
            raise
        except asyncio.TimeoutError:
            # some exchanges do not allow cancels with the client/user order id
            # so log a warning and wait for the creation of the order to complete
            self.logger().warning(
                f"Failed to cancel the order {order.client_order_id} because it does not have an exchange order id yet"
            )
            await self._order_tracker.process_order_not_found(order.client_order_id)
        except Exception as ex:
            if self._is_order_not_found_during_cancelation_error(cancelation_exception=ex):
                self.logger().warning(f"Failed to cancel order {order.client_order_id} (order not found)")
                await self._order_tracker.process_order_not_found(order.client_order_id)
            else:
                self.logger().error(f"Failed to cancel order {order.client_order_id}", exc_info=True)
        return None

    async def _execute_order_cancel_and_process_update(self, order: InFlightOrder) -> bool:
        cancelled = await self._place_cancel(order.client_order_id, order)
        if cancelled:
            update_timestamp = self.current_timestamp
            if update_timestamp is None or math.isnan(update_timestamp):
                update_timestamp = self._time()
            order_update: OrderUpdate = OrderUpdate(
                client_order_id=order.client_order_id,
                trading_pair=order.trading_pair,
                update_timestamp=update_timestamp,
                new_state=(OrderState.CANCELED
                           if self.is_cancel_request_in_exchange_synchronous
                           else OrderState.PENDING_CANCEL),
            )
            self._order_tracker.process_order_update(order_update)
        return cancelled

    async def _execute_cancel(self, trading_pair: str, order_id: str) -> str:
        """
        Requests the exchange to cancel an active order

        :param trading_pair: the trading pair the order to cancel operates with
        :param order_id: the client id of the order to cancel
        """
        result = None
        tracked_order = self._order_tracker.fetch_tracked_order(order_id)
        if tracked_order is not None:
            result = await self._execute_order_cancel(order=tracked_order)

        return result

    # === Order Tracking ===

    def restore_tracking_states(self, saved_states: Dict[str, Any]):
        """
        Restore in-flight orders from saved tracking states, this is st the connector can pick up on where it left off
        when it disconnects.

        :param saved_states: The saved tracking_states.
        """
        self._order_tracker.restore_tracking_states(tracking_states=saved_states)

    def start_tracking_order(self,
                             order_id: str,
                             exchange_order_id: Optional[str],
                             trading_pair: str,
                             trade_type: TradeType,
                             price: Decimal,
                             amount: Decimal,
                             order_type: OrderType,
                             **kwargs):
        """
        Starts tracking an order by adding it to the order tracker.

        :param order_id: the order identifier
        :param exchange_order_id: the identifier for the order in the exchange
        :param trading_pair: the token pair for the operation
        :param trade_type: the type of order (buy or sell)
        :param price: the price for the order
        :param amount: the amount for the order
        :param order_type: type of execution for the order (MARKET, LIMIT, LIMIT_MAKER)
        """
        self._order_tracker.start_tracking_order(
            InFlightOrder(
                client_order_id=order_id,
                exchange_order_id=exchange_order_id,
                trading_pair=trading_pair,
                order_type=order_type,
                trade_type=trade_type,
                amount=amount,
                price=price,
                creation_timestamp=self.current_timestamp
            )
        )

    def stop_tracking_order(self, order_id: str):
        """
        Stops tracking an order

        :param order_id: The id of the order that will not be tracked any more
        """
        self._order_tracker.stop_tracking_order(client_order_id=order_id)

    async def _sleep(self, delay: float):
        await asyncio.sleep(delay)

    # === Implementation-specific methods ===

    @abstractmethod
    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        raise NotImplementedError

    @abstractmethod
    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs,
                           ) -> Tuple[str, float]:
        raise NotImplementedError

    @abstractmethod
    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> AddedToCostTradeFee:
        raise NotImplementedError

    # === Network-API-related code ===

    # overridden in implementation of exchanges
    #
    web_utils = None

    async def start_network(self):
        """
        Start all required tasks to update the status of the connector. Those tasks include:
        - The order book tracker
        - The polling loops to update the trading rules and trading fees
        - The polling loop to update order status and balance status using REST API (backup for main update process)
        - The background task to process the events received through the user stream tracker (websocket connection)
        """
        await self.stop_network()
        self.order_book_tracker.start()
        if self.is_trading_required:
            self._trading_rules_polling_task = safe_ensure_future(self._trading_rules_polling_loop())
            self._trading_fees_polling_task = safe_ensure_future(self._trading_fees_polling_loop())
            self._status_polling_task = safe_ensure_future(self._status_polling_loop())
            self._user_stream_tracker_task = self._create_user_stream_tracker_task()
            self._user_stream_event_listener_task = safe_ensure_future(self._user_stream_event_listener())
            self._lost_orders_update_task = safe_ensure_future(self._lost_orders_update_polling_loop())

    async def check_network(self) -> NetworkStatus:
        """
        Checks connectivity with the exchange using the API
        """
        try:
            await self._make_network_check_request()
        except asyncio.CancelledError:
            raise
        except Exception:
            return NetworkStatus.NOT_CONNECTED
        return NetworkStatus.CONNECTED

    async def stop_network(self):
        # Resets timestamps and events for status_polling_loop
        self._last_poll_timestamp = 0
        self._last_timestamp = 0
        self._poll_notifier = asyncio.Event()

        self.order_book_tracker.stop()
        if self._status_polling_task is not None:
            self._status_polling_task.cancel()
            self._status_polling_task = None
        if self._trading_rules_polling_task is not None:
            self._trading_rules_polling_task.cancel()
            self._trading_rules_polling_task = None
        if self._trading_fees_polling_task is not None:
            self._trading_fees_polling_task.cancel()
            self._trading_fees_polling_task = None
        if self._user_stream_tracker_task is not None:
            self._user_stream_tracker_task.cancel()
            self._user_stream_tracker_task = None

        # Stop the user stream tracker to properly clean up child tasks
        if self._user_stream_tracker is not None:
            await self._user_stream_tracker.stop()
        if self._user_stream_event_listener_task is not None:
            self._user_stream_event_listener_task.cancel()
            self._user_stream_event_listener_task = None
        if self._lost_orders_update_task is not None:
            self._lost_orders_update_task.cancel()
            self._lost_orders_update_task = None

    async def add_trading_pair(self, trading_pair: str) -> bool:
        """
        Dynamically adds a trading pair to the connector.
        This method handles order book subscription and tracking.

        Subclasses (e.g., perpetual connectors) may override this to add
        additional initialization like funding info.

        :param trading_pair: the trading pair to add (e.g., "BTC-USDT")
        :return: True if successfully added, False otherwise
        """
        return await self.order_book_tracker.add_trading_pair(trading_pair)

    async def remove_trading_pair(self, trading_pair: str) -> bool:
        """
        Dynamically removes a trading pair from the connector.
        This method handles order book unsubscription and cleanup.

        Subclasses (e.g., perpetual connectors) may override this to add
        additional cleanup like funding info.

        :param trading_pair: the trading pair to remove (e.g., "BTC-USDT")
        :return: True if successfully removed, False otherwise
        """
        return await self.order_book_tracker.remove_trading_pair(trading_pair)

    # === loops and sync related methods ===
    #
    async def _trading_rules_polling_loop(self):
        """
        Updates the trading rules by requesting the latest definitions from the exchange.
        Executes regularly every 30 minutes
        """
        while True:
            try:
                await safe_gather(self._update_trading_rules())
                await self._sleep(self.TRADING_RULES_INTERVAL)
            except NotImplementedError:
                raise
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unexpected error while fetching trading rules.", exc_info=True,
                    app_warning_msg=f"Could not fetch new trading rules from {self.name_cap}"
                                    " Check network connection.")
                await self._sleep(0.5)

    async def _trading_fees_polling_loop(self):
        """
        Only some exchanges provide a fee endpoint.
        If _update_trading_fees() is not defined, we just exit the loop
        """
        while True:
            try:
                await safe_gather(self._update_trading_fees())
                await self._sleep(self.TRADING_FEES_INTERVAL)
            except NotImplementedError:
                raise
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unexpected error while fetching trading fees.", exc_info=True,
                    app_warning_msg=f"Could not fetch new trading fees from {self.name_cap}."
                                    " Check network connection.")
                await self._sleep(0.5)

    async def _status_polling_loop(self):
        """
        Performs all required operation to keep the connector updated and synchronized with the exchange.
        It contains the backup logic to update status using API requests in case the main update source
        (the user stream data source websocket) fails.
        It also updates the time synchronizer. This is necessary because the exchange requires
        the time of the client to be the same as the time in the exchange.
        Executes when the _poll_notifier event is enabled by the `tick` function.
        """
        while True:
            try:
                await self._poll_notifier.wait()
                await self._update_time_synchronizer()

                # the following method is implementation-specific
                await self._status_polling_loop_fetch_updates()

                self._last_poll_timestamp = self.current_timestamp
                self._poll_notifier = asyncio.Event()
            except asyncio.CancelledError:
                raise
            except NotImplementedError:
                raise
            except Exception:
                self.logger().network(
                    "Unexpected error while fetching account updates.",
                    exc_info=True,
                    app_warning_msg=f"Could not fetch account updates from {self.name_cap}. "
                                    "Check API key and network connection.")
                await self._sleep(0.5)

    async def _update_time_synchronizer(self, pass_on_non_cancelled_error: bool = False):
        try:
            await self._time_synchronizer.update_server_time_offset_with_time_provider(
                time_provider=self.web_utils.get_current_server_time(
                    throttler=self._throttler,
                    domain=self.domain,
                )
            )
        except asyncio.CancelledError:
            raise
        except Exception:
            if not pass_on_non_cancelled_error:
                self.logger().exception(f"Error requesting time from {self.name_cap} server")
                raise

    async def _lost_orders_update_polling_loop(self):
        """
        This loop regularly executes the update of lost orders, to keep receiving any new order fill or status change
        until we are totally sure the order is no longer alive in the exchange
        """
        while True:
            try:
                await self._cancel_lost_orders()
                await self._update_lost_orders_status()
                await self._sleep(self.SHORT_POLL_INTERVAL)
            except NotImplementedError:
                raise
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error while updating the time synchronizer")
                await self._sleep(0.5)

    async def _iter_user_event_queue(self) -> AsyncIterable[Dict[str, any]]:
        """
        Called by _user_stream_event_listener.
        """
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Error while reading user events queue. Retrying in 1s.")
                await self._sleep(1.0)

    def _is_user_stream_initialized(self):
        return self._user_stream_tracker.data_source.last_recv_time > 0 or not self.is_trading_required

    def _create_user_stream_tracker(self):
        return UserStreamTracker(data_source=self._create_user_stream_data_source())

    def _create_user_stream_tracker_task(self):
        return safe_ensure_future(self._user_stream_tracker.start())

    # === Exchange / Trading logic methods that call the API ===

    async def _update_trading_rules(self):
        exchange_info = await self._make_trading_rules_request()
        trading_rules_list = await self._format_trading_rules(exchange_info)
        self._trading_rules.clear()
        for trading_rule in trading_rules_list:
            self._trading_rules[trading_rule.trading_pair] = trading_rule
        self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)

    async def _api_get(self, *args, **kwargs):
        kwargs["method"] = RESTMethod.GET
        return await self._api_request(*args, **kwargs)

    async def _api_post(self, *args, **kwargs):
        kwargs["method"] = RESTMethod.POST
        return await self._api_request(*args, **kwargs)

    async def _api_put(self, *args, **kwargs):
        kwargs["method"] = RESTMethod.PUT
        return await self._api_request(*args, **kwargs)

    async def _api_delete(self, *args, **kwargs):
        kwargs["method"] = RESTMethod.DELETE
        return await self._api_request(*args, **kwargs)

    async def _api_request_url(self, path_url: str, is_auth_required: bool = False) -> str:
        if is_auth_required:
            url = self.web_utils.private_rest_url(path_url, domain=self.domain)
        else:
            url = self.web_utils.public_rest_url(path_url, domain=self.domain)

        return url

    async def _api_request(
            self,
            path_url,
            overwrite_url: Optional[str] = None,
            method: RESTMethod = RESTMethod.GET,
            params: Optional[Dict[str, Any]] = None,
            data: Optional[Dict[str, Any]] = None,
            is_auth_required: bool = False,
            return_err: bool = False,
            limit_id: Optional[str] = None,
            headers: Optional[Dict[str, Any]] = None,
            **kwargs,
    ) -> Dict[str, Any]:

        last_exception = None
        rest_assistant = await self._web_assistants_factory.get_rest_assistant()

        url = overwrite_url or await self._api_request_url(path_url=path_url, is_auth_required=is_auth_required)

        for _ in range(2):
            try:
                request_result = await rest_assistant.execute_request(
                    url=url,
                    params=params,
                    data=data,
                    method=method,
                    is_auth_required=is_auth_required,
                    return_err=return_err,
                    throttler_limit_id=limit_id if limit_id else path_url,
                    headers=headers,
                )

                return request_result
            except IOError as request_exception:
                last_exception = request_exception
                if self._is_request_exception_related_to_time_synchronizer(request_exception=request_exception):
                    await self._update_time_synchronizer()
                else:
                    raise

        # Failed even after the last retry
        raise last_exception

    async def _status_polling_loop_fetch_updates(self):
        """
        Called by _status_polling_loop, which executes after each tick() is executed
        """
        await safe_gather(
            self._update_all_balances(),
            self._update_order_status(),
        )

    async def _update_all_balances(self):
        try:
            await self._update_balances()
            if not self.real_time_balance_update:
                # This is only required for exchanges that do not provide balance update notifications through websocket
                self._in_flight_orders_snapshot = {k: copy.copy(v) for k, v in self.in_flight_orders.items()}
                self._in_flight_orders_snapshot_timestamp = self.current_timestamp
        except asyncio.CancelledError:
            raise
        except Exception as request_error:
            self.logger().warning(
                f"Failed to update balances. Error: {request_error}",
                exc_info=request_error,
            )

    async def _update_orders_fills(self, orders: List[InFlightOrder]):
        for order in orders:
            try:
                trade_updates = await self._all_trade_updates_for_order(order=order)
                for trade_update in trade_updates:
                    self._order_tracker.process_trade_update(trade_update)
            except asyncio.CancelledError:
                raise
            except Exception as request_error:
                self.logger().warning(
                    f"Failed to fetch trade updates for order {order.client_order_id}. Error: {request_error}",
                    exc_info=request_error,
                )

    async def _handle_update_error_for_active_order(self, order: InFlightOrder, error: Exception):
        try:
            raise error
        except asyncio.TimeoutError:
            self.logger().debug(
                f"Tracked order {order.client_order_id} does not have an exchange id. "
                f"Attempting fetch in next polling interval."
            )
            await self._order_tracker.process_order_not_found(order.client_order_id)
        except asyncio.CancelledError:
            raise
        except Exception as request_error:
            self.logger().warning(
                f"Error fetching status update for the active order {order.client_order_id}: {request_error}.",
            )
            self.logger().debug(f"Order {order.client_order_id} not found counter: {self._order_tracker._order_not_found_records.get(order.client_order_id, 0)}")
            await self._order_tracker.process_order_not_found(order.client_order_id)

    async def _handle_update_error_for_lost_order(self, order: InFlightOrder, error: Exception):
        is_not_found = self._is_order_not_found_during_status_update_error(status_update_exception=error)
        self.logger().debug(f"Order update error for lost order {order.client_order_id}\n{order}\nIs order not found: {is_not_found} ({error})")
        if is_not_found:
            self._update_order_after_failure(order.client_order_id, order.trading_pair, exception=error)
        else:
            self.logger().warning(f"Error fetching status update for the lost order {order.client_order_id}: {error}.")

    async def _update_orders_with_error_handler(self, orders: List[InFlightOrder], error_handler: Callable):
        for order in orders:
            try:
                order_update = await self._request_order_status(tracked_order=order)
                self._order_tracker.process_order_update(order_update)
            except asyncio.CancelledError:
                raise
            except Exception as request_error:
                await error_handler(order, request_error)

    async def _update_orders(self):
        orders_to_update = self.in_flight_orders.copy()
        await self._update_orders_with_error_handler(
            orders=list(orders_to_update.values()), error_handler=self._handle_update_error_for_active_order
        )

    async def _update_lost_orders(self):
        orders_to_update = self._order_tracker.lost_orders.copy()
        await self._update_orders_with_error_handler(
            orders=list(orders_to_update.values()), error_handler=self._handle_update_error_for_lost_order
        )

    async def _update_order_status(self):
        await self._update_orders_fills(orders=list(self._order_tracker.all_fillable_orders.values()))
        await self._update_orders()

    async def _update_lost_orders_status(self):
        await self._update_orders_fills(orders=list(self._order_tracker.lost_orders.values()))
        await self._update_lost_orders()

    async def _cancel_lost_orders(self):
        for _, lost_order in self._order_tracker.lost_orders.items():
            await self._execute_order_cancel(order=lost_order)

    # Methods tied to specific API data formats
    #
    @abstractmethod
    async def _update_trading_fees(self):
        raise NotImplementedError

    @abstractmethod
    async def _user_stream_event_listener(self):
        raise NotImplementedError

    @abstractmethod
    async def _format_trading_rules(self, exchange_info_dict: Dict[str, Any]) -> List[TradingRule]:
        raise NotImplementedError

    @abstractmethod
    async def _update_balances(self):
        raise NotImplementedError

    @abstractmethod
    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        raise NotImplementedError

    @abstractmethod
    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        raise NotImplementedError

    @abstractmethod
    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        raise NotImplementedError

    @abstractmethod
    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        raise NotImplementedError

    @abstractmethod
    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        raise NotImplementedError

    @abstractmethod
    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        raise NotImplementedError

    def _create_order_tracker(self) -> ClientOrderTracker:
        return ClientOrderTracker(connector=self)

    async def _initialize_trading_pair_symbol_map(self):
        try:
            exchange_info = await self._make_trading_pairs_request()
            self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)
        except Exception:
            self.logger().exception("There was an error requesting exchange info.")

    async def _make_network_check_request(self):
        await self._api_get(path_url=self.check_network_request_path)

    async def _make_trading_rules_request(self) -> Any:
        exchange_info = await self._api_get(path_url=self.trading_rules_request_path)
        return exchange_info

    async def _make_trading_pairs_request(self) -> Any:
        exchange_info = await self._api_get(path_url=self.trading_pairs_request_path)
        return exchange_info

    def _get_poll_interval(self, timestamp: float) -> float:
        last_user_stream_message_time = (
            0 if self._user_stream_tracker is None else self._user_stream_tracker.last_recv_time
        )
        last_recv_diff = timestamp - last_user_stream_message_time
        poll_interval = (
            self.SHORT_POLL_INTERVAL if last_recv_diff > self.TICK_INTERVAL_LIMIT else self.LONG_POLL_INTERVAL
        )
        return poll_interval



import logging
from abc import ABC, abstractmethod
import functools
from sqlalchemy import (
    Column,
)


@functools.total_ordering
class DatabaseTransformation(ABC):
    def __init__(self, migrator):
        self.migrator = migrator

    @abstractmethod
    def apply(self, db_handle):
        pass

    @property
    @abstractmethod
    def name(self):
        return ""

    @property
    def from_version(self):
        return 0

    @property
    @abstractmethod
    def to_version(self):
        return None

    def does_apply_to_version(self, original_version: int, target_version: int) -> bool:
        if self.to_version is not None:
            # from_version > 0 means from_version property was overridden by transformation class
            if self.from_version > 0:
                return (self.from_version >= original_version) and (self.to_version <= target_version)
            return (self.to_version > original_version) and (self.to_version <= target_version)
        return False

    def __eq__(self, other):
        return (self.to_version == other.to_version) and (self.from_version == other.to_version)

    def __lt__(self, other):
        if self.to_version == other.to_version:
            return self.from_version < other.from_version
        else:
            return self.to_version < other.to_version

    def add_column(self, engine, table_name, column: Column, dry_run=True):
        column_name = column.compile(dialect=engine.dialect)
        column_type = column.type.compile(engine.dialect)
        column_nullable = "NULL" if column.nullable else "NOT NULL"
        query_to_execute = f'ALTER TABLE \"{table_name}\" ADD COLUMN {column_name} {column_type} {column_nullable}'
        if dry_run:
            logging.getLogger().info(f"Query to execute in DB: {query_to_execute}")
        else:
            engine.execute(query_to_execute)



import asyncio
import logging
from typing import Optional

from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger


class NotifierBase:
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self):
        self._started = False
        self._message_queue: asyncio.Queue = asyncio.Queue()
        self._send_message_task: Optional[asyncio.Task] = None

    def add_message_to_queue(self, message: str):
        self._message_queue.put_nowait(message)

    def start(self):
        if self._send_message_task is None:
            self._send_message_task = safe_ensure_future(self.send_message_from_queue())

    def stop(self):
        if self._send_message_task:
            self._send_message_task.cancel()
            self._send_message_task = None

    async def send_message_from_queue(self):
        while True:
            try:
                new_msg: str = await self._message_queue.get()
                if isinstance(new_msg, str) and len(new_msg) > 0:
                    await self._send_message(new_msg)
            except Exception as e:
                self.logger().error(str(e))
            except asyncio.CancelledError:
                raise
            await self._sleep(1.0)

    async def _sleep(self, seconds: float):
        await asyncio.sleep(seconds)

    async def _send_message(self, message: str):
        """
        A
        """
        raise NotImplementedError



# distutils: language=c++
# distutils: sources=hummingbot/core/cpp/OrderBookEntry.cpp
import bisect
import logging
import time
from typing import (
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
)

import numpy as np
import pandas as pd

from cython.operator cimport(
    address as ref,
    dereference as deref,
    postincrement as inc,
)

from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_query_result import OrderBookQueryResult
from hummingbot.core.data_type.order_book_row import OrderBookRow
from hummingbot.core.data_type.OrderBookEntry cimport truncateOverlapEntries
from hummingbot.logger import HummingbotLogger
from hummingbot.core.event.events import (
    OrderBookEvent,
    OrderBookTradeEvent
)

cimport numpy as np

ob_logger = None
NaN = float("nan")


cdef class OrderBook(PubSub):
    ORDER_BOOK_TRADE_EVENT_TAG = OrderBookEvent.TradeEvent.value

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global ob_logger
        if ob_logger is None:
            ob_logger = logging.getLogger(__name__)
        return ob_logger

    def __init__(self, dex=False):
        super().__init__()
        self._snapshot_uid = 0
        self._last_diff_uid = 0
        self._best_bid = self._best_ask = float("NaN")
        self._last_trade_price = float("NaN")
        self._last_applied_trade = -1000.0
        self._last_trade_price_rest_updated = -1000
        self._dex = dex

    cdef c_apply_diffs(self, vector[OrderBookEntry] bids, vector[OrderBookEntry] asks, int64_t update_id):
        cdef:
            set[OrderBookEntry].iterator bid_book_end = self._bid_book.end()
            set[OrderBookEntry].iterator ask_book_end = self._ask_book.end()
            set[OrderBookEntry].reverse_iterator bid_iterator
            set[OrderBookEntry].iterator ask_iterator
            set[OrderBookEntry].iterator result
            OrderBookEntry top_bid
            OrderBookEntry top_ask

        # Apply the diffs. Diffs with 0 amounts mean deletion.
        for bid in bids:
            result = self._bid_book.find(bid)
            if result != bid_book_end:
                self._bid_book.erase(result)
            if bid.getAmount() > 0:
                self._bid_book.insert(bid)
        for ask in asks:
            result = self._ask_book.find(ask)
            if result != ask_book_end:
                self._ask_book.erase(result)
            if ask.getAmount() > 0:
                self._ask_book.insert(ask)

        # If any overlapping entries between the bid and ask books, centralised: newer entries win, dex: see OrderBookEntry.cpp
        truncateOverlapEntries(self._bid_book, self._ask_book, self._dex)

        # Record the current best prices, for faster c_get_price() calls.
        bid_iterator = self._bid_book.rbegin()
        ask_iterator = self._ask_book.begin()
        if bid_iterator != self._bid_book.rend():
            top_bid = deref(bid_iterator)
            self._best_bid = top_bid.getPrice()
        if ask_iterator != self._ask_book.end():
            top_ask = deref(ask_iterator)
            self._best_ask = top_ask.getPrice()

        # Remember the last diff update ID.
        self._last_diff_uid = update_id

    cdef c_apply_snapshot(self, vector[OrderBookEntry] bids, vector[OrderBookEntry] asks, int64_t update_id):
        cdef:
            double best_bid_price = float("NaN")
            double best_ask_price = float("NaN")
            set[OrderBookEntry].reverse_iterator bid_iterator
            set[OrderBookEntry].iterator ask_iterator
            OrderBookEntry top_bid
            OrderBookEntry top_ask

        # Start with an empty order book, and then insert all entries.
        self._bid_book.clear()
        self._ask_book.clear()
        for bid in bids:
            self._bid_book.insert(bid)
            if not (bid.getPrice() <= best_bid_price):
                best_bid_price = bid.getPrice()
        for ask in asks:
            self._ask_book.insert(ask)
            if not (ask.getPrice() >= best_ask_price):
                best_ask_price = ask.getPrice()

        if self._dex:
            truncateOverlapEntries(self._bid_book, self._ask_book, self._dex)
            # Record the current best prices, for faster c_get_price() calls.
            bid_iterator = self._bid_book.rbegin()
            ask_iterator = self._ask_book.begin()
            if bid_iterator != self._bid_book.rend():
                top_bid = deref(bid_iterator)
                best_bid_price = top_bid.getPrice()
            if ask_iterator != self._ask_book.end():
                top_ask = deref(ask_iterator)
                best_ask_price = top_ask.getPrice()

        # Record the current best prices, for faster c_get_price() calls.
        self._best_bid = best_bid_price
        self._best_ask = best_ask_price

        # Remember the last snapshot update ID.
        self._snapshot_uid = update_id

    cdef c_apply_trade(self, object trade_event):
        self._last_trade_price = trade_event.price
        self._last_applied_trade = time.perf_counter()
        self.c_trigger_event(self.ORDER_BOOK_TRADE_EVENT_TAG, trade_event)

    @property
    def last_trade_price(self) -> float:
        return self._last_trade_price

    @last_trade_price.setter
    def last_trade_price(self, value: float):
        self._last_trade_price = value

    @property
    def last_applied_trade(self) -> float:
        return self._last_applied_trade

    @property
    def last_trade_price_rest_updated(self) -> float:
        return self._last_trade_price_rest_updated

    @last_trade_price_rest_updated.setter
    def last_trade_price_rest_updated(self, value: float):
        self._last_trade_price_rest_updated = value

    @property
    def snapshot_uid(self) -> int:
        return self._snapshot_uid

    @property
    def last_diff_uid(self) -> int:
        return self._last_diff_uid

    @property
    def snapshot(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        bids_rows = list(self.bid_entries())
        asks_rows = list(self.ask_entries())
        bids_df = pd.DataFrame(data=bids_rows, columns=OrderBookRow._fields, dtype="float64")
        asks_df = pd.DataFrame(data=asks_rows, columns=OrderBookRow._fields, dtype="float64")
        return bids_df, asks_df

    def apply_diffs(self, bids: List[OrderBookRow], asks: List[OrderBookRow], update_id: int):
        cdef:
            vector[OrderBookEntry] cpp_bids
            vector[OrderBookEntry] cpp_asks
        for row in bids:
            cpp_bids.push_back(OrderBookEntry(row.price, row.amount, row.update_id))
        for row in asks:
            cpp_asks.push_back(OrderBookEntry(row.price, row.amount, row.update_id))
        self.c_apply_diffs(cpp_bids, cpp_asks, update_id)

    def apply_snapshot(self, bids: List[OrderBookRow], asks: List[OrderBookRow], update_id: int):
        cdef:
            vector[OrderBookEntry] cpp_bids
            vector[OrderBookEntry] cpp_asks
        for row in bids:
            cpp_bids.push_back(OrderBookEntry(row.price, row.amount, row.update_id))
        for row in asks:
            cpp_asks.push_back(OrderBookEntry(row.price, row.amount, row.update_id))
        self.c_apply_snapshot(cpp_bids, cpp_asks, update_id)

    def apply_trade(self, trade: OrderBookTradeEvent):
        self.c_apply_trade(trade)

    def apply_pandas_diffs(self, bids_df: pd.DataFrame, asks_df: pd.DataFrame):
        """
        The diffs data frame must have 3 columns, [price, amount, update_id], and a UNIX timestamp index.

        All columns are of double type.
        """
        self.apply_numpy_diffs(bids_df.values, asks_df.values)

    def apply_numpy_diffs(self, bids_array: np.ndarray, asks_array: np.ndarray):
        """
        The diffs data frame must have 3 columns, [price, amount, update_id].
        All columns are of double type.
        """
        self.c_apply_numpy_diffs(bids_array, asks_array)

    cdef c_apply_numpy_diffs(self,
                             np.ndarray[np.float64_t, ndim=2] bids_array,
                             np.ndarray[np.float64_t, ndim=2] asks_array):
        """
        The diffs data frame must have 3 columns, [price, amount, update_id].
        All columns are of double type.
        """
        cdef:
            vector[OrderBookEntry] cpp_bids
            vector[OrderBookEntry] cpp_asks
            int64_t last_update_id = 0

        for row in bids_array:
            cpp_bids.push_back(OrderBookEntry(row[0], row[1], <int64_t>(row[2])))
            last_update_id = max(last_update_id, <int64_t>row[2])
        for row in asks_array:
            cpp_asks.push_back(OrderBookEntry(row[0], row[1], <int64_t>(row[2])))
            last_update_id = max(last_update_id, <int64_t>row[2])
        self.c_apply_diffs(cpp_bids, cpp_asks, last_update_id)

    def apply_numpy_snapshot(self, bids_array: np.ndarray, asks_array: np.ndarray):
        """
        The diffs data frame must have 3 columns, [price, amount, update_id].
        All columns are of double type.
        """
        self.c_apply_numpy_snapshot(bids_array, asks_array)

    cdef c_apply_numpy_snapshot(self,
                                np.ndarray[np.float64_t, ndim=2] bids_array,
                                np.ndarray[np.float64_t, ndim=2] asks_array):
        """
        The diffs data frame must have 3 columns, [price, amount, update_id].
        All columns are of double type.
        """
        cdef:
            vector[OrderBookEntry] cpp_bids
            vector[OrderBookEntry] cpp_asks
            int64_t last_update_id = 0

        for row in bids_array:
            cpp_bids.push_back(OrderBookEntry(row[0], row[1], <int64_t>(row[2])))
            last_update_id = max(last_update_id, <int64_t>row[2])
        for row in asks_array:
            cpp_asks.push_back(OrderBookEntry(row[0], row[1], <int64_t>(row[2])))
            last_update_id = max(last_update_id, <int64_t>row[2])
        self.c_apply_snapshot(cpp_bids, cpp_asks, last_update_id)

    def bid_entries(self) -> Iterator[OrderBookRow]:
        cdef:
            set[OrderBookEntry].reverse_iterator it = self._bid_book.rbegin()
            OrderBookEntry entry
        while it != self._bid_book.rend():
            entry = deref(it)
            yield OrderBookRow(entry.getPrice(), entry.getAmount(), entry.getUpdateId())
            inc(it)

    def ask_entries(self) -> Iterator[OrderBookRow]:
        cdef:
            set[OrderBookEntry].iterator it = self._ask_book.begin()
            OrderBookEntry entry
        while it != self._ask_book.end():
            entry = deref(it)
            yield OrderBookRow(entry.getPrice(), entry.getAmount(), entry.getUpdateId())
            inc(it)

    def simulate_buy(self, amount: float) -> List[OrderBookRow]:
        amount_left = amount
        retval = []
        for ask_entry in self.ask_entries():
            ask_entry = ask_entry
            if ask_entry.amount < amount_left:
                retval.append(ask_entry)
                amount_left -= ask_entry.amount
            else:
                retval.append(OrderBookRow(ask_entry.price, amount_left, ask_entry.update_id))
                amount_left = 0.0
                break
        return retval

    def simulate_sell(self, amount: float) -> List[OrderBookRow]:
        amount_left = amount
        retval = []
        for bid_entry in self.bid_entries():
            bid_entry = bid_entry
            if bid_entry.amount < amount_left:
                retval.append(bid_entry)
                amount_left -= bid_entry.amount
            else:
                retval.append(OrderBookRow(bid_entry.price, amount_left, bid_entry.update_id))
                amount_left = 0.0
                break
        return retval

    cdef double c_get_price(self, bint is_buy) except? -1:
        cdef:
            set[OrderBookEntry] *book = ref(self._ask_book) if is_buy else ref(self._bid_book)
        if deref(book).size() < 1:
            raise EnvironmentError("Order book is empty - no price quote is possible.")
        return self._best_ask if is_buy else self._best_bid

    def get_price(self, is_buy: bool) -> float:
        return self.c_get_price(is_buy)

    cdef OrderBookQueryResult c_get_price_for_volume(self, bint is_buy, double volume):
        cdef:
            double cumulative_volume = 0
            double result_price = NaN

        if is_buy:
            for order_book_row in self.ask_entries():
                cumulative_volume += order_book_row.amount
                if cumulative_volume >= volume:
                    result_price = order_book_row.price
                    break
        else:
            for order_book_row in self.bid_entries():
                cumulative_volume += order_book_row.amount
                if cumulative_volume >= volume:
                    result_price = order_book_row.price
                    break

        return OrderBookQueryResult(NaN, volume, result_price, min(cumulative_volume, volume))

    cdef OrderBookQueryResult c_get_vwap_for_volume(self, bint is_buy, double volume):
        cdef:
            double total_cost = 0
            double total_volume = 0
            double result_vwap = NaN
        if is_buy:
            for order_book_row in self.ask_entries():
                total_cost += order_book_row.amount * order_book_row.price
                total_volume += order_book_row.amount
                if total_volume >= volume:
                    total_cost -= order_book_row.amount * order_book_row.price
                    total_volume -= order_book_row.amount
                    incremental_amount = volume - total_volume
                    total_cost += incremental_amount * order_book_row.price
                    total_volume += incremental_amount
                    result_vwap = total_cost / total_volume
                    break
        else:
            for order_book_row in self.bid_entries():
                total_cost += order_book_row.amount * order_book_row.price
                total_volume += order_book_row.amount
                if total_volume >= volume:
                    total_cost -= order_book_row.amount * order_book_row.price
                    total_volume -= order_book_row.amount
                    incremental_amount = volume - total_volume
                    total_cost += incremental_amount * order_book_row.price
                    total_volume += incremental_amount
                    result_vwap = total_cost / total_volume
                    break

        return OrderBookQueryResult(NaN, volume, result_vwap, min(total_volume, volume))

    cdef OrderBookQueryResult c_get_price_for_quote_volume(self, bint is_buy, double quote_volume):
        cdef:
            double cumulative_volume = 0
            double result_price = NaN

        if is_buy:
            for order_book_row in self.ask_entries():
                cumulative_volume += order_book_row.amount * order_book_row.price
                if cumulative_volume >= quote_volume:
                    result_price = order_book_row.price
                    break
        else:
            for order_book_row in self.bid_entries():
                cumulative_volume += order_book_row.amount * order_book_row.price
                if cumulative_volume >= quote_volume:
                    result_price = order_book_row.price
                    break

        return OrderBookQueryResult(NaN, quote_volume, result_price, min(cumulative_volume, quote_volume))

    cdef OrderBookQueryResult c_get_quote_volume_for_base_amount(self, bint is_buy, double base_amount):
        cdef:
            double cumulative_volume = 0
            double cumulative_base_amount = 0
            double row_amount = 0

        if is_buy:
            for order_book_row in self.ask_entries():
                row_amount = order_book_row.amount
                if row_amount + cumulative_base_amount >= base_amount:
                    row_amount = base_amount - cumulative_base_amount
                cumulative_base_amount += row_amount
                cumulative_volume += row_amount * order_book_row.price
                if cumulative_base_amount >= base_amount:
                    break
        else:
            for order_book_row in self.bid_entries():
                row_amount = order_book_row.amount
                if row_amount + cumulative_base_amount >= base_amount:
                    row_amount = base_amount - cumulative_base_amount
                cumulative_base_amount += row_amount
                cumulative_volume += row_amount * order_book_row.price
                if cumulative_base_amount >= base_amount:
                    break

        return OrderBookQueryResult(NaN, base_amount, NaN, cumulative_volume)

    cdef OrderBookQueryResult c_get_volume_for_price(self, bint is_buy, double price):
        cdef:
            double cumulative_volume = 0
            double result_price = NaN

        if is_buy:
            for order_book_row in self.ask_entries():
                if order_book_row.price > price:
                    break
                cumulative_volume += order_book_row.amount
                result_price = order_book_row.price
        else:
            for order_book_row in self.bid_entries():
                if order_book_row.price < price:
                    break
                cumulative_volume += order_book_row.amount
                result_price = order_book_row.price

        return OrderBookQueryResult(price, NaN, result_price, cumulative_volume)

    cdef OrderBookQueryResult c_get_quote_volume_for_price(self, bint is_buy, double price):
        cdef:
            double cumulative_volume = 0
            double result_price = NaN

        if is_buy:
            for order_book_row in self.ask_entries():
                if order_book_row.price > price:
                    break
                cumulative_volume += order_book_row.amount * order_book_row.price
                result_price = order_book_row.price
        else:
            for order_book_row in self.bid_entries():
                if order_book_row.price < price:
                    break
                cumulative_volume += order_book_row.amount * order_book_row.price
                result_price = order_book_row.price

        return OrderBookQueryResult(price, NaN, result_price, cumulative_volume)

    def get_price_for_volume(self, is_buy: bool, volume: float) -> OrderBookQueryResult:
        return self.c_get_price_for_volume(is_buy, volume)

    def get_vwap_for_volume(self, is_buy: bool, volume: float) -> OrderBookQueryResult:
        return self.c_get_vwap_for_volume(is_buy, volume)

    def get_price_for_quote_volume(self, is_buy: bool, quote_volume: float) -> OrderBookQueryResult:
        return self.c_get_price_for_quote_volume(is_buy, quote_volume)

    def get_quote_volume_for_base_amount(self, is_buy: bool, base_amount: float) -> OrderBookQueryResult:
        return self.c_get_quote_volume_for_base_amount(is_buy, base_amount)

    def get_volume_for_price(self, bint is_buy, double price) -> OrderBookQueryResult:
        return self.c_get_volume_for_price(is_buy, price)

    def get_quote_volume_for_price(self, is_buy: bool, price: float) -> OrderBookQueryResult:
        return self.c_get_quote_volume_for_price(is_buy, price)

    def restore_from_snapshot_and_diffs(self, snapshot: OrderBookMessage, diffs: List[OrderBookMessage]):
        replay_position = bisect.bisect_right(diffs, snapshot)
        replay_diffs = diffs[replay_position:]
        self.apply_snapshot(snapshot.bids, snapshot.asks, snapshot.update_id)
        for diff in replay_diffs:
            self.apply_diffs(diff.bids, diff.asks, diff.update_id)



import logging
from typing import Any, Dict, List, Optional

import hummingbot.connector.exchange.ndax.ndax_constants as CONSTANTS
from hummingbot.connector.exchange.ndax.ndax_order_book_message import NdaxOrderBookMessage
from hummingbot.core.data_type.order_book import OrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.logger import HummingbotLogger

_logger = None


class NdaxOrderBook(OrderBook):
    @classmethod
    def logger(cls) -> HummingbotLogger:
        global _logger
        if _logger is None:
            _logger = logging.getLogger(__name__)
        return _logger

    @classmethod
    def snapshot_message_from_exchange(cls,
                                       msg: Dict[str, any],
                                       timestamp: float,
                                       metadata: Optional[Dict] = None):
        """
        Convert json snapshot data into standard OrderBookMessage format
        :param msg: json snapshot data from live web socket stream
        :param timestamp: timestamp attached to incoming data
        :return: NdaxOrderBookMessage
        """

        if metadata:
            msg.update(metadata)

        return NdaxOrderBookMessage(
            message_type=OrderBookMessageType.SNAPSHOT,
            content=msg,
            timestamp=timestamp
        )

    @classmethod
    def diff_message_from_exchange(cls,
                                   msg: Dict[str, any],
                                   timestamp: Optional[float] = None,
                                   metadata: Optional[Dict] = None):
        """
        Convert json diff data into standard OrderBookMessage format
        :param msg: json diff data from live web socket stream
        :param timestamp: timestamp attached to incoming data
        :return: NdaxOrderBookMessage
        """

        if metadata:
            msg.update(metadata)

        return NdaxOrderBookMessage(
            message_type=OrderBookMessageType.DIFF,
            content=msg,
            timestamp=timestamp
        )

    @classmethod
    def trade_message_from_exchange(cls,
                                    msg: Dict[str, Any],
                                    timestamp: Optional[float] = None,
                                    metadata: Optional[Dict] = None):
        """
        Convert a trade data into standard OrderBookMessage format
        :param msg: json trade data from live web socket stream
        :param timestamp: timestamp attached to incoming data
        :return: NdaxOrderBookMessage
        """

        if metadata:
            msg.update(metadata)

        # Data fields are obtained from OrderTradeEvents
        msg.update({
            "exchange_order_id": msg.get("TradeId"),
            "trade_type": msg.get("Side"),
            "price": msg.get("Price"),
            "amount": msg.get("Quantity"),
        })

        return NdaxOrderBookMessage(
            message_type=OrderBookMessageType.TRADE,
            content=msg,
            timestamp=timestamp
        )

    @classmethod
    def from_snapshot(cls, snapshot: OrderBookMessage):
        raise NotImplementedError(CONSTANTS.EXCHANGE_NAME + " order book needs to retain individual order data.")

    @classmethod
    def restore_from_snapshot_and_diffs(cls, snapshot: OrderBookMessage, diffs: List[OrderBookMessage]):
        raise NotImplementedError(CONSTANTS.EXCHANGE_NAME + " order book needs to retain individual order data.")



import logging
from typing import Optional
import dataclasses
from hummingbot.core.event.event_listener cimport EventListener

er_logger = None

cdef class EventReporter(EventListener):
    """
    Event listener that log events to logger
    """
    def __init__(self, event_source: Optional[str] = None):
        super().__init__()
        self.event_source = event_source

    @classmethod
    def logger(cls):
        global er_logger
        if er_logger is None:
            er_logger = logging.getLogger(__name__)
        return er_logger

    cdef c_call(self, object event_object):
        try:
            if dataclasses.is_dataclass(event_object):
                event_dict = dataclasses.asdict(event_object)
            else:
                event_dict = event_object._asdict()

            event_dict.update({"event_name": event_object.__class__.__name__,
                               "event_source": self.event_source})
            self.logger().event_log(event_dict)
        except Exception:
            self.logger().error("Error logging events.", exc_info=True)



import asyncio
import math
import time
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.dexalot import dexalot_constants as CONSTANTS, dexalot_web_utils as web_utils
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.dexalot.dexalot_exchange import DexalotExchange


class DexalotAPIOrderBookDataSource(OrderBookTrackerDataSource):
    _logger: Optional[HummingbotLogger] = None
    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'DexalotExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__(trading_pairs)
        self._connector = connector
        self._domain = domain
        self._api_factory = api_factory
        self._snapshot_messages_queue_key = "order_book_snapshot"

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        pass

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            if not self._connector._evm_params:
                await self._connector._update_trading_rules()
            for trading_pair in self._trading_pairs:
                symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                min_price_increment = self._connector.trading_rules[trading_pair].min_price_increment
                show_decimal = int(-math.log10(min_price_increment))
                payload = {
                    "data": symbol,
                    "pair": symbol,
                    "type": "subscribe",
                    "decimal": show_decimal
                }
                subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
                await ws.send(subscribe_orderbook_request)

                self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...",
                exc_info=True
            )
            raise

    async def _connected_websocket_assistant(self) -> WSAssistant:
        url = f"{web_utils.wss_url(self._domain)}"

        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_timestamp: float = self._time()
        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": snapshot_timestamp,
            "bids": [],
            "asks": [],
        }
        snapshot_msg: OrderBookMessage = OrderBookMessage(
            OrderBookMessageType.SNAPSHOT,
            order_book_message_content,
            snapshot_timestamp)
        return snapshot_msg

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=raw_message["pair"])
        for trade_data in raw_message["data"]:
            timestamp = int(datetime.strptime(trade_data['ts'], '%Y-%m-%dT%H:%M:%S.%fZ').timestamp())
            trade_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.TRADE, {
                "trading_pair": trading_pair,
                "trade_type": float(TradeType.SELL.value) if trade_data["takerSide"] == 1 else float(
                    TradeType.BUY.value),
                "trade_id": trade_data["execId"],
                "price": trade_data["price"],
                "amount": trade_data["quantity"]
            }, timestamp=timestamp)

            message_queue.put_nowait(trade_message)

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        timestamp: float = time.time()

        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(
            raw_message["pair"])

        data = raw_message["data"]
        row_bids = [[price, amount] for price, amount in
                    zip(data["buyBook"][0]["prices"].split(','), data["buyBook"][0]["quantities"].split(','))]
        row_asks = [[price, amount] for price, amount in
                    zip(data["sellBook"][0]["prices"].split(','), data["sellBook"][0]["quantities"].split(','))]

        bids = [list(self._connector._format_evmamount_to_amount(trading_pair, Decimal(evm_price), Decimal(evm_amount)))
                for
                evm_price, evm_amount in row_bids]
        asks = [list(self._connector._format_evmamount_to_amount(trading_pair, Decimal(evm_price), Decimal(evm_amount)))
                for
                evm_price, evm_amount in row_asks]

        order_book_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": trading_pair,
            "update_id": timestamp,
            "bids": bids,
            "asks": asks
        }, timestamp=timestamp)
        message_queue.put_nowait(order_book_message)

    async def listen_for_order_book_diffs(self, ev_loop: asyncio.AbstractEventLoop, output: asyncio.Queue):
        """
        Suppressing call to this function as the orderbook snapshots are handled by
        listen_for_order_book_diffs() for dexalot
        """
        pass

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        stream_name = event_message.get("type")
        if stream_name == "orderBooks":
            channel = self._snapshot_messages_queue_key
        elif stream_name == "lastTrade":
            channel = self._trade_messages_queue_key
        return channel

    def _time(self):
        return time.time()

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        subscribe_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return subscribe_id

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribe to order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to subscribe to
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot subscribe: WebSocket connection not established")
            return False

        try:
            if not self._connector._evm_params:
                await self._connector._update_trading_rules()
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
            min_price_increment = self._connector.trading_rules[trading_pair].min_price_increment
            show_decimal = int(-math.log10(min_price_increment))
            payload = {
                "data": symbol,
                "pair": symbol,
                "type": "subscribe",
                "decimal": show_decimal
            }
            subscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
            await self._ws_assistant.send(subscribe_orderbook_request)

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Subscribed to public order book and trade channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred subscribing to {trading_pair}...",
                exc_info=True
            )
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribe from order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot unsubscribe: WebSocket connection not established")
            return False

        try:
            symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
            min_price_increment = self._connector.trading_rules[trading_pair].min_price_increment
            show_decimal = int(-math.log10(min_price_increment))
            payload = {
                "data": symbol,
                "pair": symbol,
                "type": "unsubscribe",
                "decimal": show_decimal
            }
            unsubscribe_orderbook_request: WSJSONRequest = WSJSONRequest(payload=payload)
            await self._ws_assistant.send(unsubscribe_orderbook_request)

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Unsubscribed from public order book and trade channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred unsubscribing from {trading_pair}...",
                exc_info=True
            )
            return False



from typing import Any, Dict

from pydantic import ConfigDict, Field, SecretStr

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.core.utils.tracking_nonce import get_tracking_nonce

CENTRALIZED = True
EXAMPLE_PAIR = "BTC-CAD"
HUMMINGBOT_ID_PREFIX = 777

# NDAX fees: https://ndax.io/fees
# Fees have to be expressed as percent value
DEFAULT_FEES = [0.2, 0.2]


# USE_ETHEREUM_WALLET not required because default value is false
# FEE_TYPE not required because default value is Percentage
# FEE_TOKEN not required because the fee is not flat

def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information
    :param exchange_info: the exchange information for a trading pair
    :return: True if the trading pair is enabled, False otherwise
    """
    return exchange_info.get("SessionStatus", "Stopped") in ["Starting", "Running"]


def get_new_client_order_id(is_buy: bool, trading_pair: str) -> str:
    ts_micro_sec: int = get_tracking_nonce()
    return f"{HUMMINGBOT_ID_PREFIX}{ts_micro_sec}"


class NdaxConfigMap(BaseConnectorConfigMap):
    connector: str = "ndax"
    ndax_uid: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your NDAX user ID (uid)",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    ndax_account_name: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter the name of the account you want to use",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    ndax_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your NDAX API key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    ndax_secret_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your NDAX secret key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="ndax")


KEYS = NdaxConfigMap.model_construct()

OTHER_DOMAINS = ["ndax_testnet"]
OTHER_DOMAINS_PARAMETER = {"ndax_testnet": "ndax_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"ndax_testnet": "BTC-CAD"}
OTHER_DOMAINS_DEFAULT_FEES = {"ndax_testnet": [0.2, 0.2]}


class NdaxTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = "ndax_testnet"
    ndax_testnet_uid: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your NDAX Testnet user ID (uid)",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    ndax_testnet_account_name: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter the name of the account you want to use",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    ndax_testnet_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your NDAX Testnet API key",
            "is_secure": True,
            "is_connect_key": True,
            "pr}mpt_on_new": True,
        }
    )
    ndax_testnet_secret_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your NDAX Testnet secret key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="ndax_testnet")


OTHER_DOMAINS_KEYS = {"ndax_testnet": NdaxTestnetConfigMap.model_construct()}



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.kucoin.com"
HEALTH_CHECK_ENDPOINT = "/api/v1/timestamp"
CANDLES_ENDPOINT = "/api/v1/market/candles"

PUBLIC_WS_DATA_PATH_URL = "/api/v1/bullet-public"

INTERVALS = bidict({
    "1s": "1s",  # Implemented for resampling to 1s from trades in quants-lab
    "1m": "1min",
    "3m": "3min",
    "5m": "5min",
    "15m": "15min",
    "30m": "30min",
    "1h": "1hour",
    "2h": "2hour",
    "4h": "4hour",
    "6h": "6hour",
    "8h": "8hour",
    "12h": "12hour",
    "1d": "1day",
    "1w": "1week",
    "1M": "1month"
})
MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1500

MAX_REQUEST = 4000
TIME_INTERVAL = 30
REQUEST_WEIGHT = "REQUEST_WEIGHT"

RATE_LIMITS = [
    RateLimit(limit_id=REQUEST_WEIGHT, limit=MAX_REQUEST, time_interval=TIME_INTERVAL),
    RateLimit(limit_id=CANDLES_ENDPOINT, limit=MAX_REQUEST, time_interval=TIME_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=3)]),
    RateLimit(limit_id=HEALTH_CHECK_ENDPOINT, limit=MAX_REQUEST, time_interval=TIME_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=3)]),
    RateLimit(limit_id=PUBLIC_WS_DATA_PATH_URL, limit=MAX_REQUEST, time_interval=TIME_INTERVAL,
              linked_limits=[LinkedLimitWeightPair(REQUEST_WEIGHT, weight=10)]),
]



import logging
from typing import List, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.data_feed.candles_feed.binance_spot_candles import constants as CONSTANTS
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.logger import HummingbotLogger


class BinanceSpotCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)

    @property
    def name(self):
        return f"binance_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        return trading_pair.replace("-", "")

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST) -> dict:
        params = {
            "symbol": self._ex_trading_pair,
            "interval": self.interval,
            "limit": limit
        }
        if end_time:
            params["endTime"] = end_time * 1000
        return params

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        return [
            [self.ensure_timestamp_in_seconds(row[0]), row[1], row[2], row[3], row[4], row[5],
             row[7], row[8], row[9], row[10]]
            for row in data
        ]

    def ws_subscription_payload(self):
        candle_params = [f"{self._ex_trading_pair.lower()}@kline_{self.interval}"]
        payload = {
            "method": "SUBSCRIBE",
            "params": candle_params,
            "id": 1
        }
        return payload

    def _parse_websocket_message(self, data: dict):
        candles_row_dict = {}
        if data is not None and data.get("e") == "kline":  # data will be None when the websocket is disconnected
            candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(data["k"]["t"])
            candles_row_dict["open"] = data["k"]["o"]
            candles_row_dict["high"] = data["k"]["h"]
            candles_row_dict["low"] = data["k"]["l"]
            candles_row_dict["close"] = data["k"]["c"]
            candles_row_dict["volume"] = data["k"]["v"]
            candles_row_dict["quote_asset_volume"] = data["k"]["q"]
            candles_row_dict["n_trades"] = data["k"]["n"]
            candles_row_dict["taker_buy_base_volume"] = data["k"]["V"]
            candles_row_dict["taker_buy_quote_volume"] = data["k"]["Q"]
            return candles_row_dict



import hashlib
import hmac
import json
from typing import Any, Dict, List, Optional

from hummingbot.connector.exchange.bitmart import bitmart_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class BitmartAuth(AuthBase):
    """
    Auth class required by BitMart API
    Learn more at https://developer-pro.bitmart.com/en/part2/auth.html
    """

    def __init__(self, api_key: str, secret_key: str, memo: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.memo = memo
        self.time_provider: TimeSynchronizer = time_provider

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.

        :param request: the request to be configured for authenticated interaction

        :return: The RESTRequest with auth information included
        """

        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(self.authentication_headers(request=request))
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. OKX does not use this
        functionality
        """
        return request  # pass-through

    def _generate_signature(self, timestamp: str, body: Optional[str] = None) -> str:
        body = body or ""
        unsigned_signature = f"{str(timestamp)}#{self.memo}#{body}"

        signature = hmac.new(
            self.secret_key.encode("utf-8"),
            unsigned_signature.encode("utf-8"),
            hashlib.sha256).hexdigest()

        return signature

    def authentication_headers(self, request: RESTRequest) -> Dict[str, Any]:
        timestamp = str(int(self.time_provider.time() * 1e3))

        params = json.dumps(request.params) if request.params is not None else request.data

        sign = self._generate_signature(timestamp=timestamp, body=params)

        header = {
            "X-BM-KEY": self.api_key,
            "X-BM-SIGN": sign,
            "X-BM-TIMESTAMP": timestamp,
            "X-BM-BROKER-ID": CONSTANTS.BROKER_ID,
        }

        return header

    def websocket_login_parameters(self) -> List[str]:
        timestamp = str(int(self.time_provider.time() * 1e3))

        return [
            self.api_key,
            timestamp,
            self._generate_signature(
                timestamp=timestamp,
                body="bitmart.WebSocket")
        ]



from dataclasses import dataclass, field
from enum import Enum
from typing import List

from hummingbot.core.api_throttler.data_types import RateLimit

# Rate limits ID
REST_CALL_RATE_LIMIT_ID = "coin_gecko_rest_rate_limit_id"


@dataclass(frozen=True)
class CoinGeckoTier:
    """Data class representing CoinGecko API tier configuration"""
    name: str  # Name used for user configuration
    header: str  # API header name to use for authentication
    base_url: str  # Base URL for the API tier
    rate_limit: int  # Calls per minute
    rate_limits: List[RateLimit] = field(default_factory=list)  # Rate limits for this tier


# API Tiers as dataclass instances with all necessary properties
PUBLIC = CoinGeckoTier(
    name="PUBLIC",
    header=None,
    base_url="https://api.coingecko.com/api/v3",
    rate_limit=10,
    rate_limits=[RateLimit(REST_CALL_RATE_LIMIT_ID, limit=10, time_interval=60)]
)

DEMO = CoinGeckoTier(
    name="DEMO",
    header="x-cg-demo-api-key",
    base_url="https://api.coingecko.com/api/v3",
    rate_limit=50,
    rate_limits=[RateLimit(REST_CALL_RATE_LIMIT_ID, limit=50, time_interval=60)]
)

PRO = CoinGeckoTier(
    name="PRO",
    header="x-cg-pro-api-key",
    base_url="https://pro-api.coingecko.com/api/v3",
    rate_limit=500,
    rate_limits=[RateLimit(REST_CALL_RATE_LIMIT_ID, limit=500, time_interval=60)]
)

# Enum for storage and selection


class CoinGeckoAPITier(Enum):
    """
    CoinGecko's Rate Limit Tiers. Based on how much money you pay them.
    """
    PUBLIC = PUBLIC
    DEMO = DEMO
    PRO = PRO


PING_REST_ENDPOINT = "/ping"
PRICES_REST_ENDPOINT = "/coins/markets"
SUPPORTED_VS_TOKENS_REST_ENDPOINT = "/simple/supported_vs_currencies"

COOLOFF_AFTER_BAN = 60.0 * 1.05

TOKEN_CATEGORIES = [
    "cryptocurrency",
    "decentralized-exchange",
    "decentralized-finance-defi",
    "smart-contract-platform",
    "stablecoins",
    "wrapped-tokens",
]



import asyncio
from enum import Enum
from typing import Any, Dict, Optional

import ujson

from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.ws_assistant import WSAssistant


class NdaxMessageType(Enum):
    REQUEST_TYPE = 0
    REPLY_TYPE = 1
    SUBSCRIBE_TO_EVENT_TYPE = 2
    EVENT = 3
    UNSUBSCRIBE_FROM_EVENT = 4
    ERROR = 5


class NdaxWebSocketAdaptor:

    _message_type_field_name = "m"
    _message_number_field_name = "i"
    _endpoint_field_name = "n"
    _payload_field_name = "o"

    """
    Auxiliary class that works as a wrapper of a low level web socket. It contains the logic to create messages
    with the format expected by NDAX
    :param websocket: The low level socket to be used to send and receive messages
    :param previous_messages_number: number of messages already sent to NDAX. This parameter is useful when the
    connection is reestablished after a communication error, and allows to keep a unique identifier for each message.
    The default previous_messages_number is 0
    """
    MESSAGE_TIMEOUT = 20.0
    PING_TIMEOUT = 5.0

    def __init__(
        self,
        websocket: WSAssistant,
        previous_messages_number: int = 0,
    ):
        self._websocket = websocket
        self._messages_counter = previous_messages_number
        self._lock = asyncio.Lock()

    @classmethod
    def endpoint_from_raw_message(cls, raw_message: str) -> str:
        message = ujson.loads(raw_message)
        return cls.endpoint_from_message(message=message)

    @classmethod
    def endpoint_from_message(cls, message: Dict[str, Any]) -> str:
        return message.get(cls._endpoint_field_name)

    @classmethod
    def payload_from_raw_message(cls, raw_message: str) -> Dict[str, Any]:
        return cls.payload_from_message(message=raw_message)

    @classmethod
    def payload_from_message(cls, message: Dict[str, Any]) -> Dict[str, Any]:
        payload = ujson.loads(message.get(cls._payload_field_name))
        return payload

    @property
    def websocket(self) -> WSAssistant:
        return self._websocket

    async def next_message_number(self):
        async with self._lock:
            self._messages_counter += 1
            next_number = self._messages_counter
        return next_number

    async def send_request(self, endpoint_name: str, payload: Dict[str, Any], limit_id: Optional[str] = None):
        message_number = await self.next_message_number()
        message = {
            self._message_type_field_name: NdaxMessageType.REQUEST_TYPE.value,
            self._message_number_field_name: message_number,
            self._endpoint_field_name: endpoint_name,
            self._payload_field_name: ujson.dumps(payload),
        }

        message_request: WSJSONRequest = WSJSONRequest(payload=message)

        await self._websocket.send(message_request)

    async def process_websocket_messages(self, queue: asyncio.Queue):
        async for ws_response in self._websocket.iter_messages():
            data = ws_response.data
            await self._process_event_message(event_message=data, queue=queue)

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if len(event_message) > 0:
            queue.put_nowait(event_message)

    async def close(self):
        if self._websocket is not None:
            await self._websocket.disconnect()

    async def disconnect(self):
        if self._websocket is not None:
            await self._websocket.disconnect()



from os.path import dirname, join, realpath
from typing import Dict

from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import Completer
from prompt_toolkit.layout import Dimension
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Float,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.widgets import Box, Button, SearchToolbar

from hummingbot.client.config.config_helpers import ClientConfigAdapter
from hummingbot.client.settings import MAXIMUM_LOG_PANE_LINE_COUNT, MAXIMUM_OUTPUT_PANE_LINE_COUNT
from hummingbot.client.tab.data_types import CommandTab
from hummingbot.client.ui.custom_widgets import CustomTextArea as TextArea, FormattedTextLexer

HEADER = """
                                                *,.
                                                *,,,*
                                            ,,,,,,,               *
                                            ,,,,,,,,            ,,,,
                                            *,,,,,,,,(        .,,,,,,
                                        /,,,,,,,,,,     .*,,,,,,,,
                                        .,,,,,,,,,,,.  ,,,,,,,,,,,*
                                        ,,,,,,,,,,,,,,,,,,,,,,,,,,,
                            //      ,,,,,,,,,,,,,,,,,,,,,,,,,,,,#*%
                        .,,,,,,,,. *,,,,,,,,,,,,,,,,,,,,,,,,,,,%%%%%%&@
                        ,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,%%%%%%%&
                    ,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,%%%%%%%&
                    /*,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,(((((%%&
                **.         #,,,,,,,,,,,,,,,,,,,,,,,,,,,,,((((((((((#.
            **               *,,,,,,,,,,,,,,,,,,,,,,,,**/(((((((((((((*
                                ,,,,,,,,,,,,,,,,,,,,*********((((((((((((
                                ,,,,,,,,,,,,,,,**************((((((((@
                                (,,,,,,,,,,,,,,,***************(#
                                    *,,,,,,,,,,,,,,,,**************/
                                    ,,,,,,,,,,,,,,,***************/
                                        ,,,,,,,,,,,,,,****************
                                        .,,,,,,,,,,,,**************/
                                            ,,,,,,,,*******,
                                            *,,,,,,,,********
                                            ,,,,,,,,,/******/
                                            ,,,,,,,,,@  /****/
                                            ,,,,,,,,
                                            , */


██   ██ ██    ██ ███    ███ ███    ███ ██ ███    ██  ██████  ██████   ██████  ████████
██   ██ ██    ██ ████  ████ ████  ████ ██ ████   ██ ██       ██   ██ ██    ██    ██
███████ ██    ██ ██ ████ ██ ██ ████ ██ ██ ██ ██  ██ ██   ███ ██████  ██    ██    ██
██   ██ ██    ██ ██  ██  ██ ██  ██  ██ ██ ██  ██ ██ ██    ██ ██   ██ ██    ██    ██
██   ██  ██████  ██      ██ ██      ██ ██ ██   ████  ██████  ██████   ██████     ██

======================================================================================
Hummingbot is an open source software client that helps you build and run
market making, arbitrage, and other high-frequency trading bots.

- Official repo: https://github.com/hummingbot/hummingbot
- Join the community: https://discord.gg/hummingbot
- Learn market making: https://hummingbot.org/botcamp

Useful Commands:
- connect     List available exchanges and add API keys to them
- balance     See your exchange balances
- start       Start a script or strategy
- help        List all commands

"""

with open(realpath(join(dirname(__file__), '../../VERSION'))) as version_file:
    version = version_file.read().strip()


def create_input_field(lexer=None, completer: Completer = None):
    return TextArea(
        height=10,
        prompt='>>> ',
        style='class:input_field',
        multiline=False,
        focus_on_click=True,
        lexer=lexer,
        auto_suggest=AutoSuggestFromHistory(),
        completer=completer,
        complete_while_typing=True,
    )


def create_output_field(client_config_map: ClientConfigAdapter):
    return TextArea(
        style='class:output_field',
        focus_on_click=False,
        read_only=False,
        scrollbar=True,
        max_line_count=MAXIMUM_OUTPUT_PANE_LINE_COUNT,
        initial_text=HEADER,
        lexer=FormattedTextLexer(client_config_map)
    )


def create_timer():
    return TextArea(
        style='class:footer',
        focus_on_click=False,
        read_only=False,
        scrollbar=False,
        max_line_count=1,
        width=30,
    )


def create_process_monitor():
    return TextArea(
        style='class:footer',
        focus_on_click=False,
        read_only=False,
        scrollbar=False,
        max_line_count=1,
        align=WindowAlign.RIGHT
    )


def create_trade_monitor():
    return TextArea(
        style='class:footer',
        focus_on_click=False,
        read_only=False,
        scrollbar=False,
        max_line_count=1,
    )


def create_search_field() -> SearchToolbar:
    return SearchToolbar(text_if_not_searching=[('class:primary', "[CTRL + F] to start searching.")],
                         forward_search_prompt=[('class:primary', "Search logs [Press CTRL + F to hide search] >>> ")],
                         ignore_case=True)


def create_log_field(search_field: SearchToolbar):
    return TextArea(
        style='class:log_field',
        text="Running Logs\n",
        focus_on_click=False,
        read_only=False,
        scrollbar=True,
        max_line_count=MAXIMUM_LOG_PANE_LINE_COUNT,
        initial_text="Running Logs \n",
        search_field=search_field,
        preview_search=False,
    )


def create_live_field():
    return TextArea(
        style='class:log_field',
        focus_on_click=False,
        read_only=False,
        scrollbar=True,
        max_line_count=MAXIMUM_OUTPUT_PANE_LINE_COUNT,
    )


def create_log_toggle(function):
    return Button(
        text='> Ctrl+T',
        width=10,
        handler=function,
        left_symbol='',
        right_symbol='',
    )


def create_tab_button(text, function, margin=2, left_symbol=' ', right_symbol=' '):
    return Button(
        text=text,
        width=len(text) + margin,
        handler=function,
        left_symbol=left_symbol,
        right_symbol=right_symbol
    )


def get_version():
    return [("class:header", f"Version: {version}")]


def get_active_strategy():
    from hummingbot.client.hummingbot_application import HummingbotApplication
    hb = HummingbotApplication.main_application()
    style = "class:log_field"
    return [(style, f"Strategy: {hb.strategy_name}")]


def get_strategy_file():
    from hummingbot.client.hummingbot_application import HummingbotApplication
    hb = HummingbotApplication.main_application()
    style = "class:log_field"
    return [(style, f"Strategy File: {hb.strategy_file_name}")]


def get_gateway_status():
    from hummingbot.client.hummingbot_application import HummingbotApplication
    hb = HummingbotApplication.main_application()
    gateway_status = hb.trading_core.gateway_monitor.gateway_status.name
    style = "class:log_field"

    # Check if SSL is enabled
    use_ssl = getattr(hb.client_config_map.gateway, "gateway_use_ssl", False)
    lock_icon = "🔒 " if use_ssl else ""

    # Add visual indicator based on status
    if gateway_status == "ONLINE":
        status_display = f"🟢 {gateway_status}"
    else:
        status_display = f"🔴 {gateway_status}"

    return [(style, f"{lock_icon}Gateway: {status_display}")]


def generate_layout(input_field: TextArea,
                    output_field: TextArea,
                    log_field: TextArea,
                    right_pane_toggle: Button,
                    log_field_button: Button,
                    search_field: SearchToolbar,
                    timer: TextArea,
                    process_monitor: TextArea,
                    trade_monitor: TextArea,
                    command_tabs: Dict[str, CommandTab],
                    ):
    components = {}

    components["item_top_version"] = Window(FormattedTextControl(get_version), style="class:header")
    components["item_top_active"] = Window(FormattedTextControl(get_active_strategy), style="class:header")
    components["item_top_file"] = Window(FormattedTextControl(get_strategy_file), style="class:header")
    components["item_top_gateway"] = Window(FormattedTextControl(get_gateway_status), style="class:header")
    components["item_top_toggle"] = right_pane_toggle
    components["pane_top"] = VSplit([components["item_top_version"],
                                     components["item_top_active"],
                                     components["item_top_file"],
                                     components["item_top_gateway"],
                                     components["item_top_toggle"]], height=1)
    components["pane_bottom"] = VSplit([trade_monitor,
                                        process_monitor,
                                        timer], height=1)
    output_pane = Box(body=output_field, padding=0, padding_left=2, style="class:output_field")
    input_pane = Box(body=input_field, padding=0, padding_left=2, padding_top=1, style="class:input_field")
    components["pane_left"] = HSplit([output_pane, input_pane], width=Dimension(weight=1))
    if all(not t.is_selected for t in command_tabs.values()):
        log_field_button.window.style = "class:tab_button.focused"
    else:
        log_field_button.window.style = "class:tab_button"
    tab_buttons = [log_field_button]
    for tab in sorted(command_tabs.values(), key=lambda x: x.tab_index):
        if tab.button is not None:
            if tab.is_selected:
                tab.button.window.style = "class:tab_button.focused"
            else:
                tab.button.window.style = "class:tab_button"
            tab.close_button.window.style = tab.button.window.style
            tab_buttons.append(VSplit([tab.button, tab.close_button]))
    pane_right_field = log_field
    focused_right_field = [tab.output_field for tab in command_tabs.values() if tab.is_selected]
    if focused_right_field:
        pane_right_field = focused_right_field[0]
    components["pane_right_top"] = VSplit(tab_buttons, height=1, style="class:log_field", padding_char=" ", padding=2)
    components["pane_right"] = ConditionalContainer(
        Box(body=HSplit([components["pane_right_top"], pane_right_field, search_field], width=Dimension(weight=1)),
            padding=0, padding_left=2, style="class:log_field"),
        filter=True
    )
    components["hint_menus"] = [Float(xcursor=True,
                                      ycursor=True,
                                      transparent=True,
                                      content=CompletionsMenu(max_height=16,
                                                              scroll_offset=1))]

    root_container = HSplit([
        components["pane_top"],
        VSplit(
            [FloatContainer(components["pane_left"], components["hint_menus"]),
             components["pane_right"]]),
        components["pane_bottom"],
    ])
    return Layout(root_container, focused_element=input_field), components



import asyncio
import logging
import time
from typing import Any, Dict, List, Optional

from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.web_assistant.connections.data_types import WSPlainTextRequest
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.data_feed.candles_feed.bitget_perpetual_candles import constants as CONSTANTS
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.logger import HummingbotLogger


class BitgetPerpetualCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)

        self._ping_task: Optional[asyncio.Task] = None

    @property
    def name(self):
        return f"bitget_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return True

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return True

    @staticmethod
    def product_type_associated_to_trading_pair(trading_pair: str) -> str:
        """
        Returns the product type associated with the trading pair
        """
        _, quote = split_hb_trading_pair(trading_pair)

        if quote == "USDT":
            return CONSTANTS.USDT_PRODUCT_TYPE
        if quote == "USDC":
            return CONSTANTS.USDC_PRODUCT_TYPE
        if quote == "USD":
            return CONSTANTS.USD_PRODUCT_TYPE

        raise ValueError(f"No product type associated to {trading_pair} tranding pair")

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(
            url=self.health_check_url,
            throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT
        )

        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        return trading_pair.replace("-", "")

    def _get_rest_candles_params(
        self,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        limit: Optional[int] = CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST
    ) -> dict:

        params = {
            "symbol": self._ex_trading_pair,
            "productType": self.product_type_associated_to_trading_pair(self._trading_pair),
            "granularity": CONSTANTS.INTERVALS[self.interval],
            "limit": limit
        }

        if start_time is not None and end_time is not None:
            now = int(time.time())
            max_days = CONSTANTS.INTERVAL_LIMITS_DAYS.get(self.interval)

            if max_days is not None:
                allowed_seconds = max_days * 24 * 60 * 60
                earliest_allowed = now - allowed_seconds

                if start_time < earliest_allowed:
                    self.logger().error(
                        f"[Bitget API] Invalid start time for interval '{self.interval}': "
                        f"the earliest allowed start time is {earliest_allowed} "
                        f"({max_days} days before now), but requested {start_time}."
                    )
                    raise ValueError('Invalid start time for current interval. See logs for more details.')

        if start_time is not None:
            params["startTime"] = start_time * 1000
        if end_time is not None:
            params["endTime"] = end_time * 1000

        return params

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        """
        Rest response example:
        {
            "code": "00000",
            "msg": "success",
            "requestTime": 1695865615662,
            "data": [
                [
                    "1695835800000",  # Timestamp ms
                    "26210.5",        # Entry
                    "26210.5",        # Highest
                    "26194.5",        # Lowest
                    "26194.5",        # Exit
                    "26.26",          # Volume base
                    "687897.63"       # Volume quote
                ]
            ]
        }
        """
        if data and data.get("data"):
            candles = data["data"]

            return [
                [
                    self.ensure_timestamp_in_seconds(int(row[0])),
                    float(row[1]), float(row[2]), float(row[3]),
                    float(row[4]), float(row[5]), float(row[6]),
                    0., 0., 0.
                ]
                for row in candles
            ]

        return []

    def ws_subscription_payload(self):
        interval = CONSTANTS.INTERVALS[self.interval]
        channel = f"{CONSTANTS.WS_CANDLES_ENDPOINT}{interval}"
        payload = {
            "op": "subscribe",
            "args": [
                {
                    "instType": self.product_type_associated_to_trading_pair(self._trading_pair),
                    "channel": channel,
                    "instId": self._ex_trading_pair
                }
            ]
        }

        return payload

    def _parse_websocket_message(self, data: dict) -> Optional[Dict[str, Any]]:
        """
        WS response example:
        {
            "action": "snapshot",  # or "update"
            "arg": {
                "instType": "USDT-FUTURES",
                "channel": "candle1m",
                "instId": "BTCUSDT"
            },
            "data": [
                [
                    "1695835800000",  # Timestamp ms
                    "26210.5",        # Opening
                    "26210.5",        # Highest
                    "26194.5",        # Lowest
                    "26194.5",        # Closing
                    "26.26",          # Volume coin
                    "687897.63"       # Volume quote
                    "687897.63"       # Volume USDT
                ]
            ],
            "ts": 1695702747821
            }
        """
        if data == "pong":
            return

        candles_row_dict: Dict[str, Any] = {}

        if data and data.get("data") and data["action"] == "update":
            candle = data["data"][0]
            candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(int(candle[0]))
            candles_row_dict["open"] = float(candle[1])
            candles_row_dict["high"] = float(candle[2])
            candles_row_dict["low"] = float(candle[3])
            candles_row_dict["close"] = float(candle[4])
            candles_row_dict["volume"] = float(candle[5])
            candles_row_dict["quote_asset_volume"] = float(candle[6])
            candles_row_dict["n_trades"] = 0.
            candles_row_dict["taker_buy_base_volume"] = 0.
            candles_row_dict["taker_buy_quote_volume"] = 0.

            return candles_row_dict

    async def _send_ping(self, websocket_assistant: WSAssistant) -> None:
        ping_request = WSPlainTextRequest(CONSTANTS.PUBLIC_WS_PING_REQUEST)

        await websocket_assistant.send(ping_request)

    async def send_interval_ping(self, websocket_assistant: WSAssistant) -> None:
        """
        Coroutine to send PING messages periodically.

        :param websocket_assistant: The websocket assistant to use to send the PING message.
        """
        try:
            while True:
                await self._send_ping(websocket_assistant)
                await asyncio.sleep(CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        except asyncio.CancelledError:
            self.logger().info("Interval PING task cancelled")
            raise
        except Exception:
            self.logger().exception("Error sending interval PING")

    async def listen_for_subscriptions(self):
        """
        Connects to the candlestick websocket endpoint and listens to the messages sent by the
        exchange.
        """
        ws: Optional[WSAssistant] = None
        while True:
            try:
                ws: WSAssistant = await self._connected_websocket_assistant()
                await self._subscribe_channels(ws)
                self._ping_task = asyncio.create_task(self.send_interval_ping(ws))
                await self._process_websocket_messages(websocket_assistant=ws)
            except asyncio.CancelledError:
                raise
            except ConnectionError as connection_exception:
                self.logger().warning(f"The websocket connection was closed ({connection_exception})")
            except Exception:
                self.logger().exception(
                    "Unexpected error occurred when listening to public klines. Retrying in 1 seconds...",
                )
                await self._sleep(1.0)
            finally:
                if self._ping_task is not None:
                    self._ping_task.cancel()
                    try:
                        await self._ping_task
                    except asyncio.CancelledError:
                        pass
                    self._ping_task = None
                await self._on_order_stream_interruption(websocket_assistant=ws)



import asyncio
import time
import unittest
from test.isolated_asyncio_wrapper_test_case import IsolatedAsyncioWrapperTestCase, LocalClassEventLoopWrapperTestCase


class TestLocalClassEventLoopWrapperTestCase(unittest.TestCase):
    def setUp(self):
        self.test_case = LocalClassEventLoopWrapperTestCase()
        self.test_case.setUpClass()

    def tearDown(self):
        self.test_case.tearDownClass()

    def test_setUpClass_with_existing_loop(self):
        self.main_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.main_loop)

        LocalClassEventLoopWrapperTestCase.setUpClass()
        self.assertIsNotNone(LocalClassEventLoopWrapperTestCase.main_event_loop)
        self.assertEqual(self.main_loop, LocalClassEventLoopWrapperTestCase.main_event_loop)
        self.assertNotEqual(self.main_loop, LocalClassEventLoopWrapperTestCase.local_event_loop)

        self.main_loop = None

    def test_setUpClass_without_existing_loop(self):
        # Close the main event loop if it exists
        asyncio.set_event_loop(None)

        # Call setUpClass and verify that it does not create a new event loop
        LocalClassEventLoopWrapperTestCase.setUpClass()
        self.assertIsNotNone(LocalClassEventLoopWrapperTestCase.main_event_loop)
        self.assertEqual(asyncio.get_event_loop(), LocalClassEventLoopWrapperTestCase.local_event_loop)

    def test_tearDownClass_with_existing_loop(self):
        self.main_loop = asyncio.new_event_loop()

        LocalClassEventLoopWrapperTestCase.main_event_loop = self.main_loop
        LocalClassEventLoopWrapperTestCase.local_event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(LocalClassEventLoopWrapperTestCase.local_event_loop)

        LocalClassEventLoopWrapperTestCase.tearDownClass()
        self.assertIsNone(LocalClassEventLoopWrapperTestCase.main_event_loop)
        self.assertIsNone(LocalClassEventLoopWrapperTestCase.local_event_loop)
        self.assertEqual(self.main_loop, asyncio.get_event_loop())

        self.main_loop.close()
        asyncio.set_event_loop(None)
        self.main_loop = None

    def test_exception_in_test_case(self):
        IsolatedAsyncioWrapperTestCase.setUpClass()

        try:
            # Simulate a test case that raises an exception
            raise Exception("Test exception")
        except Exception:
            pass

        # Despite the exception, tearDownClass should still correctly clean up the event loop
        LocalClassEventLoopWrapperTestCase.tearDownClass()
        self.assertIsNone(LocalClassEventLoopWrapperTestCase.main_event_loop)

    def test_run_async_with_timeout(self):
        self.test_case.setUp()

        # Test a coroutine that finishes before the timeout
        start_time = time.time()
        result = self.test_case.run_async_with_timeout(asyncio.sleep(0.1), timeout=1.0)
        end_time = time.time()
        self.assertIsNone(result)
        self.assertLess(end_time - start_time, 1.0)

        # Test a coroutine that doesn't finish before the timeout
        with self.assertRaises(asyncio.TimeoutError):
            self.test_case.run_async_with_timeout(asyncio.sleep(2.0), timeout=1.0)

        self.test_case.tearDown()


if __name__ == "__main__":
    unittest.main()
